import cv2


def center_crop(img, dim):
    """Returns center cropped image
    Args:
    img: image to be center cropped
    dim: dimensions (width, height) to be cropped
    """
    width, height = img.shape[1], img.shape[0]

    # process crop width and height for max available dimension
    crop_width = dim[0] if dim[0] < img.shape[1] else img.shape[1]
    crop_height = dim[1] if dim[1] < img.shape[0] else img.shape[0]
    mid_x, mid_y = int(width / 2), int(height / 2)
    cw2, ch2 = int(crop_width / 2), int(crop_height / 2)
    crop_img = img[mid_y - ch2:mid_y + ch2, mid_x - cw2:mid_x + cw2]
    return crop_img


def scale_image(img, factor=1):
    """Returns resize image by scale factor.
    This helps to retain resolution ratio while resizing.
    Args:
    img: image to be scaled
    factor: scale factor to resize
    """
    return cv2.resize(img, (int(img.shape[1] * factor), int(img.shape[0] * factor)))



from .gaussian import Normal
from .ou import OUNoise
from .nonoise import NoNoise

def get_noise(name):
    """Returns noise model class from name."""
    if name == "Gauss":
        return Normal
    elif name == "OU":
        return OUNoise
    elif name == "NoNoise":
        return  NoNoise
    else:
        raise ValueError("Specified Noise not found!")


from typing import Tuple

import gym
import torch
import torch.nn as nn
from torch.nn.functional import one_hot

from pytorchrl.agent.actors.distributions import get_dist
from pytorchrl.agent.actors.utils import Scale, Unscale


# TODO: review prediction sizes when reward is and is not predicted!

class WorldModel(nn.Module):
    """
    Model-Based Actor class for Model-Based algorithms.

    It contains the dynamics network to predict the next state (and reward if selected). 

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    hidden_size : int
        Hidden size number.
    standard_scaler : StandardScaler
        StandardScaler class instance.
    reward_function : func
        Reward function to be learned.
    """

    def __init__(self,
                 device,
                 input_space,
                 action_space,
                 standard_scaler,
                 hidden_size=64,
                 reward_function=None,
                 ) -> None:

        super(WorldModel, self).__init__()

        self.device = device
        self.input_space = input_space
        self.action_space = action_space
        self.reward_function = reward_function

        if reward_function is not None:
            self.predict = self.predict_given_reward
        else:
            self.predict = self.predict_learned_reward

        # Scaler for scaling training inputs
        self.standard_scaler = standard_scaler
        self.hidden_size = hidden_size
        self.model = self.create_dynamics()

    def create_dynamics(self):
        """
        Create a dynamics model and define it as class attribute under the name `name`.

        Parameters
        ----------
        name : str
            dynamics model name.
        """
        if type(self.action_space) == gym.spaces.discrete.Discrete:
            input_layer = nn.Linear(self.input_space.shape[0] + self.action_space.n, out_features=self.hidden_size)
        else:
            input_layer = nn.Linear(self.input_space.shape[0] + self.action_space.shape[0], out_features=self.hidden_size)

        hidden_layer = nn.Linear(self.hidden_size, self.hidden_size)

        if self.reward_function is None:
            num_outputs = self.input_space.shape[0] + 1
        else:
            num_outputs = self.input_space.shape[0]

        output_layer = get_dist("DeterministicMB")(num_inputs=self.hidden_size, num_outputs=num_outputs)
        self.activation = nn.ReLU
        dynamics_layers = [input_layer, self.activation(), hidden_layer, self.activation(), output_layer]

        if type(self.action_space) == gym.spaces.box.Box:
            self.scale = Scale(self.action_space)
            self.unscale = Unscale(self.action_space)
        else:
            self.scale = None
            self.unscale = None

        dynamics_net = nn.Sequential(*dynamics_layers)

        return dynamics_net

    @staticmethod
    def check_dynamics_weights(parameter1, parameter2):
        for p1, p2 in zip(parameter1, parameter2):
            if p1.data.ne(p2.data).sum() > 0:
                return False
        return True

    def reinitialize_dynamics_model(self):
        """
        Re-initializes the dynamics model, can be done before each new Model learning run.
        Might help in some environments to overcome over-fitting of the model!
        """
        old_weights = self.model.parameters()
        self.create_dynamics()
        self.model.to(self.device)
        new_weights = self.model.parameters()
        assert not self.check_dynamics_weights(old_weights, new_weights)

    def predict_learned_reward(self, states: torch.Tensor, actions: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Does the next state prediction and reward prediction with a learn reward function.

        Parameters
        ----------
        states : torch.Tensor
            Current state s
        actions : torch.Tensor
            Action taken in state s

        Returns
        -------
        next_states : torch.Tensor
            Next states.
        rewards : torch.Tensor
            Reward prediction.
        """

        if type(self.action_space) == gym.spaces.discrete.Discrete:
            actions = one_hot(actions, num_classes=self.action_space.n).squeeze(1)

        inputs = torch.cat((states, actions), dim=-1)

        # scale inputs based on recent batch scalings
        norm_inputs, _ = self.standard_scaler.transform(inputs)
        norm_predictions = self.model(norm_inputs)

        # inverse transform outputs
        predictions = self.standard_scaler.inverse_transform(norm_predictions)
        predictions[:, :-1] += states.to(self.device)

        next_states = predictions[:, :-1]
        rewards = predictions[:, -1].unsqueeze(-1)

        # TODO: add Termination function?

        return next_states, rewards

    def predict_given_reward(self, states: torch.Tensor, actions: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Does the next state prediction and calculates the reward given a reward function. 

        Parameters
        ----------
        states : torch.Tensor
            Current state s
        actions : torch.Tensor
            Action taken in state s

        Returns
        -------
        next_states : torch.Tensor
            Next states.
        rewards : torch.Tensor
            Calculated reward.
        """
        if type(self.action_space) == gym.spaces.discrete.Discrete:
            actions = one_hot(actions, num_classes=self.action_space.n).squeeze(1)

        inputs = torch.cat((states, actions), dim=-1)

        # scale inputs based on recent batch scalings
        norm_inputs, _ = self.standard_scaler.transform(inputs)

        norm_predictions = self.model(norm_inputs)

        # inverse transform outputs
        predictions = self.standard_scaler.inverse_transform(norm_predictions)
        predictions += states.to(self.device)
        next_states = predictions

        rewards = self.reward_function(states, actions, next_states)

        # TODO: add Termination function?

        return next_states, rewards




import numpy as np
import torch

import pytorchrl as prl
from pytorchrl.agent.storages.base import Storage as S


def dim0_reshape(tensor, size):
    """
    Reshapes tensor so indices are defined like this:

    00, 01, 02, 03, 04, 05, 06, 07, 08, 09, size + 1, ..., self.max_size
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, size + 1, ..., self.max_size
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, size + 1, ..., self.max_size

    """
    return np.moveaxis(tensor, [0, 1], [1, 0])[:, 0: size].reshape(-1, *tensor.shape[2:])


class ReplayBuffer(S):
    """
    Storage class for Off-Policy algorithms.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device : torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance
    envs : VecEnv
        Vector of environments instance.
    """

    # Data fields to store in buffer and contained in the generated batches
    storage_tensors = prl.OffPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs):

        self.actor = actor
        self.device = device
        self.algo = algorithm
        self.recurrent_actor = actor.is_recurrent
        self.max_size, self.size, self.step = size, 0, 0
        self.data = {}  # lazy init

        if self.recurrent_actor:
            self.sequence_length = algorithm.update_every
            self.overlap_length = int(actor.sequence_overlap * self.sequence_length)
            self.non_overlap_length = self.sequence_length - self.overlap_length
            self.max_size = (self.max_size // self.sequence_length) * self.sequence_length
            self.max_size *= 2  # to account for consecutive overlapping of sequences
            self.staged_overlap = None

        self.reset()

    @classmethod
    def create_factory(cls, size):
        """
        Returns a function that creates ReplayBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.

        Returns
        -------
        create_buffer_instance : func
            creates a new ReplayBuffer class instance.
        """

        def create_buffer(device, actor, algorithm, envs):
            """Create and return a ReplayBuffer instance."""
            return cls(size, device, actor, algorithm, envs)

        return create_buffer

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        for k, v in sample.items():

            if k not in self.storage_tensors or v is None:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                size = 1
            else:
                size = self.max_size

            if isinstance(v, dict):
                self.data[k] = {}
                for x, y in sample[k].items():
                    self.data[k][x] = np.zeros((size, *y.shape), dtype=np.float32)
            else:
                self.data[k] = np.zeros((size, *v.shape), dtype=np.float32)

    def get_data_slice(self, start_pos, end_pos):
        """
        Makes a copy of all tensors in the bufer between steps `start_pos`
        and `end_pos`.

        Parameters
        ----------
        start_pos : int
            initial slice position.
        end_pos : int
            final slice position.

        Returns
        -------
        data : dict
            data slice copied from the buffer.
        """

        copied_data = {k: None for k in self.storage_tensors}

        for k, v in self.data.items():

            if v is None:
                continue

            if isinstance(self.data[k], dict):
                copied_data[k] = {x: None for x in self.data[k]}
                for x, y in v.items():
                    copied_data[k][x] = np.copy(y[start_pos:end_pos])
            else:
                copied_data[k] = np.copy(v[start_pos:end_pos])

        return copied_data

    def get_all_buffer_data(self, data_to_cpu=False):
        """
        Return all currently stored data. If data_to_cpu, no need to do
        anything since data tensors are already in cpu memory.

        Parameters
        ----------
        data_to_cpu : bool
            Whether or not to move data tensors to cpu memory.

        Returns
        -------
        data : dict
            data currently stored in the buffer.
        """

        # Define data structure
        data = {k: None if not isinstance(self.data[k], dict) else
            {x: None for x in self.data[k]} for k in self.data}

        # If recurrent, get only whole sequences
        if self.recurrent_actor:
            idx = int((self.step // self.sequence_length) * self.sequence_length)
        else:
            idx = self.step

        # Fill up data
        for k, v in self.data.items():

            if v is None:
                continue

            if isinstance(self.data[k], dict):
                for x, y in self.data[k].items():
                    data[k][x] = y[:idx]
            else:
                data[k] = v[:idx]

        # If self.actor uses RNNs, save ending of last sequence
        if self.recurrent_actor and data_to_cpu:
            self.staged_overlap = self.get_data_slice(
                self.step - self.overlap_length, self.step)

        return data

    def reset(self):
        """
        Set class size and step to zero. If self.actor uses RNNs, add overlap
        slice of last sequence before reset at the beginning of the storage.
        """

        self.size -= self.step
        self.step = 0

        if self.recurrent_actor and self.staged_overlap:
            self.insert_data_slice(self.staged_overlap)

    def insert_data_slice(self, new_data):
        """
        Appends new_data to currently stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to be added to self.data.
        """

        lengths = []
        for k, v in new_data.items():

            if v is None:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                continue

            if isinstance(new_data[k], dict):

                if self.data[k] is None:
                    self.data[k] = {i: None for i in new_data[k].keys()}

                for x, y in new_data[k].items():
                    length = self.insert_single_tensor_slice(self.data[k], x, y)
                    lengths.append(length)
            else:
                length = self.insert_single_tensor_slice(self.data, k, v)
                lengths.append(length)

        assert len(set(lengths)) == 1

        self.step = (self.step + length) % self.max_size
        self.size = min(self.size + length, self.max_size)

    def insert_single_tensor_slice(self, tensor_storage, tensor_key, tensor_values):
        """
        Appends tensor_value to buffer dict using tensor_key as key.

        Parameters
        ----------
        tensor_storage :

        tensor_key : str
            key to use to store the tensor.
        tensor_values : np.ndarray
            tensor values.

        Returns
        -------
        l : int
            length (time axe) of the tensor added to the buffer.
        """

        l = tensor_values.shape[0]

        if tensor_storage[tensor_key] is None:  # If not defined, initialize tensor
            tensor_storage[tensor_key] = np.zeros((self.max_size, *tensor_values.shape[1:]), dtype=np.float32)

        if self.step + l <= self.max_size:  # If enough space, add tensor at the end
            tensor_storage[tensor_key][self.step:self.step + l] = tensor_values

        else:  # Circular buffer
            tensor_storage[tensor_key][
            self.step:self.max_size] = tensor_values[0:self.max_size - self.step]
            tensor_storage[tensor_key][0:l - self.max_size + self.step] = tensor_values[self.max_size - self.step:]

        return l

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        # Data tensors lazy initialization
        if self.size == 0 and prl.OBS not in self.data.keys():
            self.init_tensors(sample)

        # If using memory, save fixed length consecutive overlapping sequences
        if self.recurrent_actor and self.step % self.sequence_length == 0 and self.step != 0 and self.overlap_length > 0:
            next_seq_overlap = self.get_data_slice(self.step - self.overlap_length, self.step)
            self.insert_data_slice(next_seq_overlap)

        # Insert
        for k, v in sample.items():

            if k not in self.storage_tensors:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                continue

            if isinstance(sample[k], dict):
                for x, y in sample[k].items():
                    self.data[k][x][self.step] = y.cpu()
            else:
                self.data[k][self.step] = v.cpu()

        # Update
        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def before_gradients(self):
        """
        Steps required before updating actor policy model.
        """
        pass

    def after_gradients(self, batch, info):
        """
        Steps required after updating actor policy model

        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """
        return info

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.

        Yields
        ------
        batch : dict
            Generated data batches.
        """

        num_proc = self.data[prl.DONE].shape[1]

        for _ in range(num_mini_batch):

            if self.recurrent_actor:  # Batches to feed a recurrent actor

                sequences_x_batch = mini_batch_size // self.sequence_length + 1

                assert self.size % self.sequence_length == 0, \
                    "Buffer does not contain an integer number of complete rollout sequences"

                # Define batch structure
                batch = {k: [] if not isinstance(self.data[k], dict) else
                    {x: [] for x in self.data[k]} for k in self.data.keys()}

                # Randomly select sequences
                seq_idxs = np.random.randint(0, num_proc * int(
                    self.size / self.sequence_length), size=sequences_x_batch)

                # Get data indexes
                idxs = []
                for idx in seq_idxs:
                    idxs += range(idx * self.sequence_length, (idx + 1) * self.sequence_length)

                # Fill up batch with sequences data
                for k, v in self.data.items():
                    # Only first recurrent state in each sequence needed
                    positions = seq_idxs * self.sequence_length if k in (prl.RHS, prl.RHS2) else idxs
                    if isinstance(v, dict):
                        for x, y in v.items():
                            t = dim0_reshape(y, self.size)[positions]
                            batch[k][x] = torch.as_tensor(t, dtype=torch.float32).to(self.device)
                    else:
                        t = dim0_reshape(v, self.size)[positions]
                        batch[k] = torch.as_tensor(t, dtype=torch.float32).to(self.device)
                yield batch

            else:
                batch = {k: {} for k in self.data.keys()}
                samples = np.random.randint(0, num_proc * self.size, size=mini_batch_size)
                for k, v in self.data.items():

                    if k in (prl.RHS, prl.RHS2):
                        size, idxs = 1, np.array([0])
                    else:
                        size, idxs = self.size, samples

                    if isinstance(v, dict):
                        for x, y in v.items():
                            batch[k][x] = torch.as_tensor(y[0:size].reshape(
                                -1, *y.shape[2:])[idxs], dtype=torch.float32).to(self.device)
                    else:
                        batch[k] = torch.as_tensor(v[0:size].reshape(
                            -1, *v.shape[2:])[idxs], dtype=torch.float32).to(self.device)
                yield batch

    def update_storage_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            if parameter_name == "max_size" and self.recurrent_actor:
                new_parameter_value = (new_parameter_value // self.sequence_length) * self.sequence_length
                new_parameter_value *= 2
            setattr(self, parameter_name, new_parameter_value)




import json
import pandas
import os.path as osp
from glob import glob


def load_baselines_results(dir):
    monitor_files = (
        glob(osp.join(dir, "*monitor.json")) +
        glob(osp.join(dir, "*monitor.csv"))) # get both csv and (old) json files
    if not monitor_files:
        raise ValueError("no monitor files of the form *.csv found in %s" % (dir))
    dfs = []
    headers = []
    for fname in monitor_files:
        with open(fname, 'rt') as fh:
            if fname.endswith('csv'):
                firstline = fh.readline()
                if not firstline:
                    continue
                assert firstline[0] == '#'
                header = json.loads(firstline[1:])
                df = pandas.read_csv(fh, index_col=None)
                headers.append(header)
            elif fname.endswith('json'): # Deprecated json format
                episodes = []
                lines = fh.readlines()
                header = json.loads(lines[0])
                headers.append(header)
                for line in lines[1:]:
                    episode = json.loads(line)
                    episodes.append(episode)
                df = pandas.DataFrame(episodes)
            else:
                assert 0, 'unreachable'
            df['t'] += header['t_start']
        dfs.append(df)
    df = pandas.concat(dfs)
    df.sort_values('t', inplace=True)
    df.reset_index(inplace=True)
    df['t'] -= min(header['t_start'] for header in headers)
    return df


import os
import gym
import habitat
from pytorchrl.envs.common import FrameSkip, FrameStack
from pytorchrl.envs.habitat.wrappers import NavRLEnv, MyHabitat


def habitat_env_factory(conf_file, index_worker=0, index_env=0, frame_skip=0, frame_stack=1):
    """
    Create train Obstacle Tower Unity3D environment.
    Useful info_keywords 'floor', 'start', 'seed'.
    Parameters
    ----------
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    Returns
    -------
    env : gym.Env
        Train environment.
    """
    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    config = habitat.get_config(conf_file)
    env = NavRLEnv(config)
    env.seed(config['SEED'] + index_worker + index_env)

    env = MyHabitat(env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    return env



import os
import copy
import inspect
from pytorchrl.agent.env.openai_baselines_dependencies.Monitor import Monitor
from pytorchrl.agent.env.env_wrappers import TransposeImagesIfRequired


def make_env(env_fn, env_kwargs, index_col_worker, index_grad_worker, index_env, log_dir=None, info_keywords=(), mode="train"):
    """
    Returns a function that handles the creating of a single environment, so it
    can be executed in an independent thread.

    Parameters
    ----------

    env_fn : func
        Function to create the environment.
    env_kwargs : dict
        keyword arguments of env_fn.
    log_dir : str
        Target path for bench.Monitor logger values.
    info_keywords : tuple
        Information keywords to be logged stored by bench.Monitor.
    index_col_worker:
        Index of the data collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    mode : str
        "train" or "test"

    Returns
    -------
    _thunk : func
        A function to create and return the environment. It also sets up the
        Monitor logging and used a TransposeImage wrapper if environment obs
        are images.

    """

    if log_dir:
        path = os.path.join(log_dir, "monitor_logs", mode)
        os.makedirs(path, exist_ok=True)

    # index_worker and index_env added as paramaters
    env_kwargs = copy.deepcopy(env_kwargs)
    if "index_col_worker" in inspect.getfullargspec(env_fn).args:
        env_kwargs["index_col_worker"] = index_col_worker
    if "index_grad_worker" in inspect.getfullargspec(env_fn).args:
        env_kwargs["index_grad_worker"] = index_grad_worker
    if "index_env" in inspect.getfullargspec(env_fn).args:
        env_kwargs["index_env"] = index_env

    def _thunk():
        """Creates and returns environment"""

        # Create single environment
        env = env_fn(**env_kwargs)

        # Monitor provided info_keywords
        if log_dir is not None:
            env = Monitor(
                env, os.path.join(path, "{}_{}_{}".format(
                    index_grad_worker, index_col_worker, str(index_env))),
                allow_early_resets=True, info_keywords=info_keywords)

        # if obs are images with shape (W,H,C), transpose to (C,W,H) for PyTorch convolutions
        env = TransposeImagesIfRequired(env, op=[2, 0, 1])

        return env

    return _thunk



import copy 
import random 
import torch


class OUNoise:
    """Ornstein-Uhlenbeck process."""

    def __init__(self, size, mu=0., theta=0.15, sigma=0.2):
        """Initialize parameters and noise process."""
        self.mu = mu * torch.ones(size)
        self.theta = theta
        self.sigma = sigma
        self.reset()

    def reset(self):
        """Reset the internal state (= noise) to mean (mu)."""
        self.state = copy.copy(self.mu)

    def sample(self):
        """Update internal state and return it as a noise sample."""
        x = self.state
        dx = self.theta * (self.mu - x) + self.sigma * torch.tensor([random.random() for i in range(len(x))])
        self.state = x + dx
        return self.state



import torch
import numpy as np
from collections import deque
import pytorchrl as prl
from pytorchrl.agent.storages.off_policy.replay_buffer import ReplayBuffer as S


def dim0_reshape(tensor, size):
    """
    Reshapes tensor so indices are defined like this:

    00, 01, 02, 03, 04, 05, 06, 07, 08, 09, size + 1, ..., self.max_size
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, size + 1, ..., self.max_size
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, size + 1, ..., self.max_size

    """
    return np.moveaxis(tensor, [0, 1], [1, 0])[:, 0: size].reshape(-1, *tensor.shape[2:])


class NStepReplayBuffer(S):
    """
    Storage class for Off-Policy with multi step learning (https://arxiv.org/abs/1710.02298).

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device : torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    n_step : int or float
        Number of future steps used to computed the truncated n-step return value.
    """

    # Data fields to store in buffer and contained in the generated batches
    storage_tensors = prl.OffPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs, n_step=1):

        algorithm._update_every += n_step - 1

        super(NStepReplayBuffer, self).__init__(
            size=size,  device=device, actor=actor, algorithm=algorithm, envs=envs)

        self.n_step = n_step
        self.gamma = algorithm.gamma
        self.n_step_buffer = {k: deque(maxlen=n_step) for k in self.storage_tensors}

    @classmethod
    def create_factory(cls, size, n_step=1):
        """
        Returns a function that creates NStepReplayBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        n_step : int or float
            Number of future steps used to computed the truncated n-step return value.

        Returns
        -------
        create_buffer_instance : func
            creates a new NStepReplayBuffer class instance.
        """
        def create_buffer(device, actor, algorithm, envs):
            """Create and return a NStepReplayBuffer instance."""
            return cls(size, device, actor, algorithm, envs, n_step)
        return create_buffer

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        # Data tensors lazy initialization
        if self.size == 0 and prl.OBS not in self.data.keys():
            self.init_tensors(sample)

        # If using memory, save fixed length consecutive overlapping sequences
        if self.recurrent_actor and self.step % self.sequence_length == 0 and self.step != 0:
            next_seq_overlap = self.get_data_slice(self.step - self.overlap_length, self.step)
            self.insert_data_slice(next_seq_overlap)

        # Add obs, rhs, done, act and rew to n_step buffer
        self.n_step_buffer[prl.OBS].append(sample[prl.OBS])
        self.n_step_buffer[prl.REW].append(sample[prl.REW])
        self.n_step_buffer[prl.ACT].append(sample[prl.ACT])
        self.n_step_buffer[prl.RHS].append(sample[prl.RHS])
        self.n_step_buffer[prl.DONE].append(sample[prl.DONE])

        if len(self.n_step_buffer[prl.OBS]) == self.n_step:

            # Add obs2, rhs2 and done2 directly
            for k in (prl.OBS2, prl.RHS2, prl.DONE2):

                if not self.recurrent_actor and k == prl.RHS2:
                    continue

                if isinstance(sample[k], dict):
                    for x, y in sample[k].items():
                        self.data[k][x][self.step] = y.cpu()
                else:
                    self.data[k][self.step] = sample[k].cpu()

            # Compute done and rew
            (self.data[prl.REW][self.step],
             self.data[prl.DONE][self.step]) = self._nstep_return()

            # Get obs, rhs and act from step buffer
            for k in (prl.OBS, prl.RHS, prl.ACT):

                if not self.recurrent_actor and k == prl.RHS:
                    continue

                tensor = self.n_step_buffer[k].popleft()
                if isinstance(tensor, dict):
                    for x, y in tensor.items():
                        self.data[k][x][self.step] = y.cpu()
                else:
                    self.data[k][self.step] = tensor.cpu()

            # Update
            self.step = (self.step + 1) % self.max_size
            self.size = min(self.size + 1, self.max_size)

    def _nstep_return(self):
        """
        Computes truncated n-step returns.

        Returns
        -------
        ret : numpy.ndarray
            Next sample returns, to store in buffer.
        done : numpy.ndarray
            Next sample dones, to store in buffer.
        """

        ret = self.n_step_buffer[prl.REW][self.n_step - 1].clone()
        done = self.n_step_buffer[prl.DONE][self.n_step - 1].clone()
        for i in reversed(range(self.n_step - 1)):
            ret = ret * self.gamma * (1 - self.n_step_buffer[prl.DONE][i + 1])\
                  + self.n_step_buffer[prl.REW][i]
            done = done + self.n_step_buffer[prl.DONE][i]

        self.n_step_buffer[prl.REW].popleft()
        self.n_step_buffer[prl.DONE].popleft()

        return ret.cpu(), done.cpu()

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.

        Yields
        ------
        batch : dict
            Generated data batches.
        """

        num_proc = self.data[prl.DONE].shape[1]

        for _ in range(num_mini_batch):

            if self.recurrent_actor:  # Batches to feed a recurrent actor

                sequences_x_batch = mini_batch_size // self.sequence_length + 1

                assert self.size % self.sequence_length == 0, \
                    "Buffer does not contain an integer number of complete rollout sequences"

                # Define batch structure
                batch = {k: [] if not isinstance(self.data[k], dict) else
                    {x: [] for x in self.data[k]} for k in self.data.keys()}

                # Randomly select sequences
                seq_idxs = np.random.randint(0, num_proc * int(
                    self.size / self.sequence_length), size=sequences_x_batch)

                # Get data indexes
                idxs = []
                for idx in seq_idxs:
                    idxs += range(idx * self.sequence_length, (idx + 1) * self.sequence_length)

                # Fill up batch with data
                for k, v in self.data.items():
                    # Only first recurrent state in each sequence needed
                    positions = seq_idxs * self.sequence_length if k in (prl.RHS, prl.RHS2) else idxs
                    if isinstance(v, dict):
                        for x, y in v.items():
                            t = dim0_reshape(y, self.size)[positions]
                            batch[k][x] = torch.as_tensor(t, dtype=torch.float32).to(self.device)
                    else:
                        t = dim0_reshape(v, self.size)[positions]
                        batch[k] = torch.as_tensor(t, dtype=torch.float32).to(self.device)

                batch.update({"n_step": self.n_step})
                yield batch

            else:
                batch = {k: {} for k in self.data.keys()}
                samples = np.random.randint(0, num_proc * self.size, size=mini_batch_size)
                for k, v in self.data.items():

                    if k in (prl.RHS, prl.RHS2):
                        size, idxs = 1, np.array([0])
                    else:
                        size, idxs = self.size, samples

                    if isinstance(v, dict):
                        for x, y in v.items():
                            batch[k][x] = torch.as_tensor(y[0:size].reshape(
                                -1, *y.shape[2:])[idxs], dtype=torch.float32).to(self.device)
                    else:
                        batch[k] = torch.as_tensor(v[0:size].reshape(
                            -1, *v.shape[2:])[idxs], dtype=torch.float32).to(self.device)

                batch.update({"n_step": self.n_step})
                yield batch

    def update_storage_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            if parameter_name == "max_size" and self.recurrent_actor:
                new_parameter_value = (new_parameter_value // self.sequence_length) * self.sequence_length
                new_parameter_value *= 2
                new_parameter_value += self.n_step
            setattr(self, parameter_name, new_parameter_value)



import torch
import torch.nn as nn
import torch.optim as optim

import pytorchrl as prl
from pytorchrl.utils import RunningMeanStd, clip_grad_norm_
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients
from pytorchrl.agent.actors.feature_extractors import default_feature_extractor


class RND_PPO(Algorithm):
    """
    Exploration by Random Network Distillation with Proximal Policy Optimization algorithm class.

    Algorithm class to execute RND PPO, from Burda et al., 2018
    (https://arxiv.org/abs/1810.12894). Algorithms are modules generally
    required by multiple workers, so RND_PPO.algo_factory(...) returns a function
    that can be passed on to workers to instantiate their own RND_PPO module.

    Parameters
    ----------
    device: torch.device
        CPU or specific GPU where class computations will take place.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    lr : float
        Optimizer learning rate.
    eps : float
        Optimizer epsilon parameter.
    num_epochs : int
        Number of PPO epochs.
    gamma : float
        Discount factor parameter.
    clip_param : float
        PPO clipping parameter.
    num_mini_batch : int
        Number of batches to create from collected data for actor updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations.
    max_grad_norm : float
        Gradient clipping parameter.
    entropy_coef : float
        PPO entropy coefficient parameter.
    value_loss_coef : float
        PPO value coefficient parameter.
    use_clipped_value_loss : bool
        Prevent value loss from shifting too fast.
    policy_loss_addons : list
        List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.
    gamma_intrinsic: float
        Discount factor parameter for intrinsic rewards.
    ext_adv_coeff: float
        Extrinsic advantage coefficient.
    int_adv_coeff: float
        Intrinsic advantage coefficient.
    predictor_proportion: float
        Proportion of buffer sample to use to train the predictor network.
    pre_normalization_steps: int
        Number of obs running average normalization steps to take before starting to train.
    pre_normalization_length: int
        Length of each pre normalization steps (in environment steps).
    intrinsic_rewards_network : nn.Module
        PyTorch nn.Module used for target and predictor networks.
    intrinsic_rewards_target_network_kwargs : dict
        Keyword arguments for the target network.
    intrinsic_rewards_predictor_network_kwargs : dict
        Keyword arguments for the predictor network.

    Examples
    --------
    >>> create_algo = RND_PPO.create_factory(
        lr=0.01, eps=1e-5, num_epochs=4, clip_param=0.2,
        entropy_coef=0.01, value_loss_coef=0.5, max_grad_norm=0.5,
        num_mini_batch=4, use_clipped_value_loss=True, gamma=0.99)
    """

    def __init__(self,
                 envs,
                 actor,
                 device,
                 lr=1e-4,
                 eps=1e-8,
                 gamma=0.99,
                 num_epochs=4,
                 clip_param=0.2,
                 num_mini_batch=1,
                 test_every=1000,
                 max_grad_norm=2.0,
                 entropy_coef=0.01,
                 value_loss_coef=0.5,
                 num_test_episodes=5,
                 gamma_intrinsic=0.99,
                 ext_adv_coeff=2.0,
                 int_adv_coeff=1.0,
                 predictor_proportion=2.0,
                 pre_normalization_steps=50,
                 pre_normalization_length=128,
                 use_clipped_value_loss=False,
                 intrinsic_rewards_network=None,
                 intrinsic_rewards_target_network_kwargs={},
                 intrinsic_rewards_predictor_network_kwargs={},
                 policy_loss_addons=[]):

        # ---- General algo attributes ----------------------------------------

        # Discount factor
        self._gamma = gamma

        # Number of steps collected with initial random policy
        self._start_steps = 0  # Default to 0 for On-policy algos

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(num_epochs)

        # Number of data samples collected between network update stages
        self._update_every = None  # Depends on storage capacity

        # Number mini batches per epoch
        self._num_mini_batch = int(num_mini_batch)

        # Size of update mini batches
        self._mini_batch_size = None  # Depends on storage capacity

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = int(num_test_episodes)

        # ---- RND-PPO-specific attributes ----------------------------------------

        self.envs = envs
        self.actor = actor
        self.device = device
        self.clip_param = clip_param
        self.entropy_coef = entropy_coef
        self.max_grad_norm = max_grad_norm
        self.value_loss_coef = value_loss_coef
        self.use_clipped_value_loss = use_clipped_value_loss

        self.ext_adv_coeff = ext_adv_coeff
        self.int_adv_coeff = int_adv_coeff
        self.gamma_intrinsic = gamma_intrinsic
        self.predictor_proportion = predictor_proportion
        self.pre_normalization_steps = pre_normalization_steps
        self.pre_normalization_length = pre_normalization_length

        assert hasattr(self.actor, "value_net1"), "RND_PPO requires value critic"
        assert hasattr(self.actor, "ivalue_net1"), "RND_PPO requires ivalue critic"

        # Get observation shape
        obs_space = self.envs.observation_space.shape

        # Get frame stack value
        frame_stack = 1
        if "frame_stack" in self.envs.env_kwargs.keys():
            frame_stack = self.envs.env_kwargs["frame_stack"]

        # Get number of obs channels
        self.obs_channels = int(obs_space[0] / frame_stack)

        # Define network type
        int_net = intrinsic_rewards_network or default_feature_extractor(self.envs.observation_space)

        # Create target model
        setattr(
            self.actor, "rnd_target_net",
            int_net((self.obs_channels,) + obs_space[1:],
                    **intrinsic_rewards_target_network_kwargs).to(self.device))

        # Freeze target model parameters
        for param in self.actor.rnd_target_net.parameters():
            param.requires_grad = False

        # Create predictor model
        setattr(
            self.actor, "rnd_predictor_net",
            int_net((self.obs_channels,) + obs_space[1:],
                    **intrinsic_rewards_predictor_network_kwargs).to(self.device))

        # Define running means for int reward and obs
        self.state_rms = RunningMeanStd(shape=(1, ) + obs_space[1:], device=self.device)

        print("---Pre_normalization started.---")
        obs, rhs, done = self.actor.actor_initial_states(envs.reset())
        total_obs = torch.zeros(
            (self.pre_normalization_length,  obs.shape[0], self.obs_channels) + obs.shape[2:]).to(self.device)
        for i in range(self.pre_normalization_steps * self.pre_normalization_length):
            _, clipped_action, rhs, _ = self.acting_step(obs, rhs, done)
            obs, _, _, _ = envs.step(clipped_action)
            total_obs[i % self.pre_normalization_length].copy_(obs[:, -self.obs_channels:, ...])
            if i % self.pre_normalization_length == 0 and i != 0:
                self.state_rms.update(total_obs.reshape(-1, *total_obs.shape[2:]))
                print("{}/{}".format(i//self.pre_normalization_length, self.pre_normalization_steps))
        envs.reset()
        print("---Pre_normalization is done.---")

        # ----- Policy Loss Addons --------------------------------------------

        # Sanity check, policy_loss_addons is a PolicyLossAddOn instance
        # or a list of PolicyLossAddOn instances
        assert isinstance(policy_loss_addons, (PolicyLossAddOn, list)),\
            "RND PPO policy_loss_addons parameter should be a  PolicyLossAddOn instance " \
            "or a list of PolicyLossAddOn instances"
        if isinstance(policy_loss_addons, list):
            for addon in policy_loss_addons:
                assert isinstance(addon, PolicyLossAddOn), \
                    "RND PPO policy_loss_addons parameter should be a  PolicyLossAddOn " \
                    "instance or a list of PolicyLossAddOn instances"
        else:
            policy_loss_addons = [policy_loss_addons]

        self.policy_loss_addons = policy_loss_addons
        for addon in self.policy_loss_addons:
            addon.setup(self.actor, self.device)

        # ----- Optimizers ----------------------------------------------------

        self.optimizer = optim.Adam(self.actor.parameters(), lr=lr, eps=eps)

    @classmethod
    def create_factory(cls,
                       lr=1e-4,
                       eps=1e-8,
                       gamma=0.99,
                       num_epochs=4,
                       clip_param=0.2,
                       num_mini_batch=1,
                       test_every=1000,
                       max_grad_norm=0.5,
                       entropy_coef=0.01,
                       value_loss_coef=0.5,
                       num_test_episodes=5,
                       gamma_intrinsic=0.99,
                       ext_adv_coeff=2.0,
                       int_adv_coeff=1.0,
                       predictor_proportion=2.0,
                       pre_normalization_steps=50,
                       pre_normalization_length=128,
                       use_clipped_value_loss=True,
                       intrinsic_rewards_network=None,
                       intrinsic_rewards_target_network_kwargs={},
                       intrinsic_rewards_predictor_network_kwargs={},
                       policy_loss_addons=[]):
        """
        Returns a function to create new RND PPO instances.

        Parameters
        ----------
        lr : float
            Optimizer learning rate.
        eps : float
            Optimizer epsilon parameter.
        num_epochs : int
            Number of PPO epochs.
        gamma : float
            Discount factor parameter.
        clip_param : float
            PPO clipping parameter.
        num_mini_batch : int
            Number of batches to create from collected data for actor update.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations.
        max_grad_norm : float
            Gradient clipping parameter.
        entropy_coef : float
            PPO entropy coefficient parameter.
        value_loss_coef : float
            PPO value coefficient parameter.
        use_clipped_value_loss : bool
            Prevent value loss from shifting too fast.
        gamma_intrinsic: float
            Discount factor parameter for intrinsic rewards.
        ext_adv_coeff: float
            Extrinsic advantage coefficient.
        int_adv_coeff: float
            Intrinsic advantage coefficient.
        predictor_proportion: float
            Proportion of buffer sample to use to train the predictor network.
        pre_normalization_steps: int
            Number of obs running average normalization steps to take before starting to train.
        pre_normalization_length: int
            Length of each pre normalization steps (in environment steps).
        intrinsic_rewards_network : nn.Module
            PyTorch nn.Module used for target and predictor networks.
        intrinsic_rewards_target_network_kwargs : dict
            Keyword arguments for the target network.
        intrinsic_rewards_predictor_network_kwargs : dict
            Keyword arguments for the predictor network.
        policy_loss_addons : list
            List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new PPO class instance.
        algo_name : str
            Name of the algorithm.
        """
        def create_algo_instance(device, actor, envs):
            return cls(lr=lr,
                       eps=eps,
                       envs=envs,
                       actor=actor,
                       gamma=gamma,
                       device=device,
                       test_every=test_every,
                       num_epochs=num_epochs,
                       clip_param=clip_param,
                       entropy_coef=entropy_coef,
                       ext_adv_coeff=ext_adv_coeff,
                       int_adv_coeff=int_adv_coeff,
                       max_grad_norm=max_grad_norm,
                       num_mini_batch=num_mini_batch,
                       value_loss_coef=value_loss_coef,
                       gamma_intrinsic=gamma_intrinsic,
                       num_test_episodes=num_test_episodes,
                       predictor_proportion=predictor_proportion,
                       pre_normalization_length=pre_normalization_length,
                       pre_normalization_steps=pre_normalization_steps,
                       use_clipped_value_loss=use_clipped_value_loss,
                       intrinsic_rewards_network=intrinsic_rewards_network,
                       intrinsic_rewards_target_network_kwargs=intrinsic_rewards_target_network_kwargs,
                       intrinsic_rewards_predictor_network_kwargs=intrinsic_rewards_predictor_network_kwargs,
                       policy_loss_addons=policy_loss_addons)

        return create_algo_instance, prl.RND_PPO

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return self._gamma

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        PPO acting function.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other: dict
            Additional PPO predictions, value score and action log probability.
        """

        with torch.no_grad():

            (action, clipped_action, logp_action, rhs,
             entropy_dist, dist) = self.actor.get_action(
                obs, rhs, done, deterministic)

            value_dict = self.actor.get_value(obs, rhs, done)
            ext_value = value_dict.pop("value_net1")
            int_value = value_dict.pop("ivalue_net1")
            rhs = value_dict.pop("rhs")

            # predict intrinsic reward
            obs = obs[:, -self.obs_channels:, ...]
            obs = torch.clamp((obs - self.state_rms.mean.float()) / (self.state_rms.var.float() ** 0.5), -5, 5)
            # obs = torch.clamp((obs - self.state_rms.mean) / (self.state_rms.var ** 0.5), -5, 5).float()
            predictor_encoded_features = self.actor.rnd_predictor_net(obs)
            target_encoded_features = self.actor.rnd_target_net(obs)
            int_reward = (predictor_encoded_features - target_encoded_features).pow(2).mean(1).unsqueeze(1)

            other = {prl.VAL: ext_value, prl.IVAL: int_value, prl.LOGP: logp_action, prl.IREW: int_reward}

        return action, clipped_action, rhs, other

    def compute_loss(self, data):
        """
        Compute PPO loss from data batch.

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute PPO loss.

        Returns
        -------
        value_loss: torch.tensor
            value term of PPO loss.
        action_loss: torch.tensor
            policy term of PPO loss.
        dist_entropy: torch.tensor
            policy term of PPO loss.
        loss: torch.tensor
            PPO loss.
        """

        o, rhs, a, old_v = data[prl.OBS], data[prl.RHS], data[prl.ACT], data[prl.VAL]
        r, d, old_logp, adv = data[prl.RET], data[prl.DONE], data[prl.LOGP], data[prl.ADV]

        # RDN PPO
        ir, old_iv, iadv = data[prl.IRET], data[prl.IVAL], data[prl.IADV]
        mask = data[prl.MASK] if prl.MASK in data.keys() else None

        advs = adv * self.ext_adv_coeff + iadv * self.int_adv_coeff

        new_logp, dist_entropy, dist = self.actor.evaluate_actions(o, rhs, d, a)

        new_vs = self.actor.get_value(o, rhs, d)
        new_v = new_vs.get("value_net1")
        new_iv = new_vs.get("ivalue_net1")

        # Policy loss
        ratio = torch.exp(new_logp - old_logp)
        surr1 = ratio * advs
        surr2 = torch.clamp(ratio, 1.0 - self.clip_param, 1.0 + self.clip_param) * advs
        action_loss = - torch.min(surr1, surr2).mean()

        # Ext value loss
        if self.use_clipped_value_loss:
            # Ext value
            value_losses = (new_v - r).pow(2)
            value_pred_clipped = old_v + (new_v - old_v).clamp(-self.clip_param, self.clip_param)
            value_losses_clipped = (value_pred_clipped - r).pow(2)
            value_loss = 0.5 * torch.max(value_losses, value_losses_clipped).mean()
            # Int value
            ivalue_losses = (new_iv - ir).pow(2)
            ivalue_pred_clipped = old_iv + (new_iv - old_iv).clamp(-self.clip_param, self.clip_param)
            ivalue_losses_clipped = (ivalue_pred_clipped - ir).pow(2)
            ivalue_loss = 0.5 * torch.max(ivalue_losses, ivalue_losses_clipped).mean()
        else:
            # Ext value
            value_loss = 0.5 * (r - new_v).pow(2).mean() * self.value_loss_coef
            # Int value
            ivalue_loss = 0.5 * (ir - new_iv).pow(2).mean() * self.value_loss_coef

        total_value_loss = value_loss + ivalue_loss

        o = o[:, -self.obs_channels:, ...]
        o = torch.clamp((o - self.state_rms.mean.float()) / (self.state_rms.var.float() ** 0.5), -5, 5)

        # Rnd loss
        encoded_target_features = self.actor.rnd_target_net(o)
        encoded_predictor_features = self.actor.rnd_predictor_net(o)
        loss = (encoded_predictor_features - encoded_target_features).pow(2).mean(-1)
        mask2 = torch.rand(loss.size(), device=self.device)
        mask2 = (mask2 <= self.predictor_proportion).float()
        mask = mask.squeeze() * mask2 if mask is not None else mask2
        rnd_loss = (mask * loss).sum() / torch.max(mask.sum(), torch.Tensor([1]).to(self.device))

        entropy_loss = self.entropy_coef * dist_entropy

        loss = total_value_loss + action_loss - entropy_loss + rnd_loss

        # Extend policy loss with addons
        addons_info = {}
        for addon in self.policy_loss_addons:
            addon_loss, addons_info = addon.compute_loss_term(data, dist, addons_info)
            loss += addon_loss

        return value_loss, ivalue_loss, action_loss, rnd_loss, entropy_loss, loss, addons_info

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        data: dict
            data batch containing all required tensors to compute PPO loss.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor gradients.
        info: dict
            Dict containing current PPO iteration information.
        """

        value_loss, ivalue_loss, action_loss, rnd_loss, dist_entropy, loss, addons_info = self.compute_loss(batch)
        self.optimizer.zero_grad()
        loss.backward()

        clip_grad_norm_(self.actor.parameters())
        # nn.utils.clip_grad_norm_(self.actor.parameters(), self.max_grad_norm)

        pi_grads = get_gradients(self.actor.policy_net, grads_to_cpu=grads_to_cpu)
        v_grads = get_gradients(self.actor.value_net1, grads_to_cpu=grads_to_cpu)
        iv_grads = get_gradients(self.actor.ivalue_net1, grads_to_cpu=grads_to_cpu)
        pred_grads = get_gradients(self.actor.rnd_predictor_net, grads_to_cpu=grads_to_cpu)

        grads = {"pi_grads": pi_grads, "v_grads": v_grads, "iv_grads": iv_grads, "pred_grads": pred_grads}

        info = {
            "loss": loss.item(),
            "value_loss": value_loss.item(),
            "ivalue_loss": ivalue_loss.item(),
            "rnd_loss": rnd_loss.item(),
            "action_loss": action_loss.item(),
            "entropy_loss": dist_entropy.item(),
            "mean_intrinsic_rewards": batch[prl.IREW].mean().cpu().item(),
            "min_intrinsic_rewards": batch[prl.IREW].min().cpu().item(),
            "max_intrinsic_rewards": batch[prl.IREW].max().cpu().item(),
        }

        info.update(addons_info)

        return grads, info

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.

        Parameters
        ----------
        gradients: list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.policy_net,
                gradients=gradients["pi_grads"], device=self.device)
            set_gradients(
                self.actor.value_net1,
                gradients=gradients["v_grads"], device=self.device)
            set_gradients(
                self.actor.ivalue_net1,
                gradients=gradients["iv_grads"], device=self.device)
            set_gradients(
                self.actor.rnd_predictor_net,
                gradients=gradients["pred_grads"], device=self.device)

        self.optimizer.step()

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights

        Parameters
        ----------
        actor_weights: dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import os
import pandas as pd
from pytorchrl.agent.env import load_baselines_results


def get_num_unique_smiles(experiment_path):

    def unique_smiles():

        # Read monitor files
        data = load_baselines_results(os.path.join(experiment_path, "monitor_logs/train"))

        # Rank monitor files by reward
        data = data.sort_values("r", ascending=False)

        # Filter data
        data = data[data["molecule"] != "invalid_smile"]
        data = data[data["molecule"].duplicated() == False]
        data = data[data["r"] > 0.4]

        # Get number of unique smiles
        num_smiles = data.shape[0]

        return num_smiles

    return unique_smiles



import cv2
import numpy as np


def imdownscale(state, target_shape=(11, 8), max_pix_value=8):
    if state.shape[::-1] == target_shape:
        resized = state
    else:
        resized = cv2.resize(state, target_shape, interpolation=cv2.INTER_AREA)
    img = ((resized / 255.0) * max_pix_value).astype(np.int8)
    return img



import logging
logger = logging.getLogger(__name__)

import pytorchrl as prl
from pytorchrl.scheme.collection.c_worker_set import CWorkerSet
from pytorchrl.scheme.gradients.g_worker_set import GWorkerSet
from pytorchrl.scheme.updates.u_worker import UWorker


class Scheme:
    """
    Class to define training schemes and handle creation and operation
    of its workers.

    Parameters
    ----------
    algo_factory : func
        A function that creates an algorithm class.
    actor_factory : func
        A function that creates a policy.
    storage_factory : func
        A function that create a rollouts storage.
    train_envs_factory : func
        A function to create train environments.
    test_envs_factory : func
        A function to create test environments.
    num_col_workers : int
        Number of data collection workers per gradient worker.
    col_workers_communication : str
        Communication coordination pattern for data collection.
    col_workers_resources : dict
        Ray resource specs for collection remote workers.
    col_preemption_thresholds : dict
        specs about minimum fraction_samples [0 - 1.0] and minimum
        fraction_workers [0 - 1.0] required in synchronous data collection.
    num_grad_workers : int
        Number of gradient workers.
    grad_workers_communication : str
        Communication coordination pattern for gradient computation workers.
    grad_workers_resources : dict
        Ray resource specs for gradient remote workers.
    local_device : str
        "cpu" or specific GPU "cuda:`number`" to use for computation.
    decentralized_update_execution : bool
        Whether the gradients are applied in the update workers (central update)
        or broadcasted to all gradient workers for a decentralized update.
    """
    def __init__(self,

                 # core
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 train_envs_factory,
                 test_envs_factory=lambda v, x, y, z: None,

                 # collection
                 num_col_workers=1,
                 col_compress_data=False,
                 col_workers_communication=prl.SYNC,
                 col_workers_resources={"num_cpus": 1, "num_gpus": 0.5},
                 col_preemption_thresholds={"fraction_samples": 1.0, "fraction_workers": 1.0},

                 # gradients
                 num_grad_workers=1,
                 grad_compress_data=False,
                 grad_workers_communication=prl.SYNC,
                 grad_workers_resources={"num_cpus": 1, "num_gpus": 0.5},

                 # update
                 local_device=None,
                 decentralized_update_execution=False,  # OMP_NUM_THREADS=1 python -m torch.distributed.launch --nproc_per_node=8 main.py
                 ):

        assert col_workers_communication in (prl.SYNC, prl.ASYNC),\
            "col_workers_communication can only be `prl.SYNC` or `prl.ASYNC`"

        assert grad_workers_communication in (prl.SYNC, prl.ASYNC),\
            "grad_workers_communication can only be `prl.SYNC` or `prl.ASYNC`"

        col_execution = prl.PARALLEL if num_col_workers > 1 else prl.CENTRAL
        grad_execution = prl.PARALLEL if num_grad_workers > 1 else prl.CENTRAL

        col_workers_factory = CWorkerSet.create_factory(

            # core modules
            algo_factory=algo_factory,
            actor_factory=actor_factory,
            storage_factory=storage_factory,
            test_envs_factory=test_envs_factory,
            train_envs_factory=train_envs_factory,

            # col specs
            num_workers=num_col_workers - 1 if num_col_workers == 1 else num_col_workers,
            col_worker_resources=col_workers_resources,
            col_fraction_samples=col_preemption_thresholds.get("fraction_samples"),
            compress_data_to_send=col_compress_data,

            # grad specs
            total_parent_workers=num_grad_workers - 1 if num_grad_workers == 1 else num_grad_workers,
        )

        grad_workers_factory = GWorkerSet.create_factory(

            # col specs
            col_execution=col_execution,
            col_communication=col_workers_communication,
            col_workers_factory=col_workers_factory,
            col_fraction_workers=col_preemption_thresholds.get("fraction_workers"),

            # grad_specs
            num_workers=num_grad_workers - 1 if num_grad_workers == 1 else num_grad_workers,
            grad_worker_resources=grad_workers_resources,
            compress_grads_to_send=grad_compress_data,
        )

        self._update_worker = UWorker(

           # col specs
            col_fraction_workers=col_preemption_thresholds.get("fraction_workers"),

            # grad specs
            grad_execution=grad_execution,
            grad_communication=grad_workers_communication,
            grad_workers_factory=grad_workers_factory,

            # update specs
            local_device=local_device,
            decentralized_update_execution=decentralized_update_execution,
        )

        logger.warning("Created training scheme.")

    def update_worker(self):
        """Returns local worker"""
        return self._update_worker

    def get_agent_components(self):
        """Returns class names for each agent component."""
        return {
            "Actor": self._update_worker.local_worker.local_worker.actor.__class__.__name__,
            "Algorithm": self._update_worker.local_worker.local_worker.algo.__class__.__name__,
            "Storage": self._update_worker.local_worker.local_worker.storage.__class__.__name__,
        }



from causal_world.envs import CausalWorld
from causal_world.task_generators import generate_task
from pytorchrl.envs.common import FrameStack, FrameSkip, DelayedReward


def causal_world_train_env_factory(task_id="general",
                                   frame_skip=0,
                                   frame_stack=1,
                                   inference=False,
                                   reward_delay=1):
    """
    Create train Animal Olympics Unity3D environment.

    Parameters
    ----------
    task_id : int
        target causal world task to train the agent.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    inference : bool
        Whether or not to render the environment in real time.

    Returns
    -------
    env : gym.Env
        Train environment.
    """
    task = generate_task(task_generator_id=task_id)
    env = CausalWorld(task=task, enable_visualization=inference)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env



from pytorchrl.envs.atari.wrappers import wrap_deepmind, make_atari, MontezumaVisitedRoomEnv,\
    MontezumaEmbeddingsEnv, ScaleRewardEnv, PitfallEmbeddingsEnv
from pytorchrl.envs.common import DelayedReward


def atari_train_env_factory(
        env_id, index_col_worker, index_grad_worker, index_env=0, seed=0, frame_stack=1, reward_delay=1,
        episodic_life=True, clip_rewards=False, max_episode_steps=4500, sticky_actions=False,
        embeddings_shape=(11, 8), embeddings_num_values=8, use_domain_knowledge=False,
        domain_knowledge_embedding="default", double_state=False):
    """
    Create train Atari environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    episodic_life : bool
        Whether or not simulate end of episode when losing a life.
    clip_rewards : bool
        Whether or not to clip rewards between -1 and 1.
    max_episode_steps : int
        Maximum number of steps per episode.
    sticky_actions : bool
        Randomly repeat last action with probability 0.25.
    embeddings_shape : tuple
        Shape of atari embeddings (if embedding wrappers are used).
    embeddings_num_values : int
        Number of values for atari embeddings (if embedding wrappers are used).
    use_domain_knowledge : bool
        Whether or not to create embeddings using domain knowledge.
    domain_knowledge_embedding : str
        Type of domain knowledge embedding
    double_state : boo,
        Whether or not to concatenate last 2 different embeddings.

    Returns
    -------
    env : gym.Env
        Train environment.
    """
    env = make_atari(env_id, max_episode_steps=max_episode_steps, sticky_actions=sticky_actions)
    env.seed(index_grad_worker * 1000 + 100 * index_col_worker + index_env + seed)

    env = wrap_deepmind(
        env, episode_life=episodic_life,
        clip_rewards=clip_rewards,
        scale=False,
        frame_stack=frame_stack)

    if env_id == "MontezumaRevengeNoFrameskip-v4":
        env = MontezumaVisitedRoomEnv(env, 3)
        env = MontezumaEmbeddingsEnv(
            env, embeddings_shape, embeddings_num_values, use_domain_knowledge,
            domain_knowledge_embedding, double_state)
    elif env_id == "PitfallNoFrameskip-v4":
        env = PitfallEmbeddingsEnv(
            env, embeddings_shape, embeddings_num_values,
            use_domain_knowledge, double_state)
        # env = ScaleRewardEnv(env, 0.001)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env


def atari_test_env_factory(
        env_id, index_col_worker, index_grad_worker, index_env=0, seed=0, frame_stack=1,
        reward_delay=1, episodic_life=False, clip_rewards=False, max_episode_steps=4500, sticky_actions=False):
    """
    Create test Atari environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    max_episode_steps : int
        Maximum number of steps per episode.
    sticky_actions : bool
        Randomly repeat last action with probability 0.25.

    Returns
    -------
    env : gym.Env
        Test environment.
    """
    env = make_atari(env_id, max_episode_steps=max_episode_steps, sticky_actions=sticky_actions)
    env.seed(index_grad_worker * 1000 + 100 * index_col_worker + index_env + seed)
    env = wrap_deepmind(
        env, episode_life=episodic_life,
        clip_rewards=clip_rewards,
        scale=False,
        frame_stack=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env



from pytorchrl.agent.algorithms.policy_loss_addons.base import PolicyLossAddOn
from pytorchrl.agent.algorithms.policy_loss_addons.kv_similarity import AttractionKL
from pytorchrl.agent.algorithms.policy_loss_addons.kv_similarity import RepulsionKL
from pytorchrl.agent.algorithms.policy_loss_addons.reward_predictor import RewardPredictor
from pytorchrl.agent.algorithms.policy_loss_addons.return_predictor import ReturnPredictor



import gym
import numpy as np
from typing import Optional
import habitat
from habitat import Config, Dataset
from pytorchrl.envs.habitat.utils import center_crop, scale_image


class NavRLEnv(habitat.RLEnv):
    def __init__(self, config: Config, dataset: Optional[Dataset] = None):
        self.SUCCESS_REWARD = 2.5
        self.SLACK_REWARD = -0.001
        self.REWARD_MEASURE = 'distance_to_goal'
        self.SUCCESS_MEASURE = 'success'
        self._core_env_config = config
        self._reward_measure_name = self.REWARD_MEASURE
        self._success_measure_name = self.SUCCESS_MEASURE

        self._previous_measure = None
        self._previous_action = None
        super().__init__(self._core_env_config, dataset)

    def reset(self):
        self._previous_action = None
        observations = super().reset()
        self._previous_measure = self._env.get_metrics()[self._reward_measure_name]
        return observations

    def step(self, *args, **kwargs):
        self._previous_action = kwargs["action"]
        obs, reward, done, info = super().step(*args, **kwargs)
        return obs, reward, done, info

    def get_reward_range(self):
        return (
            self.SLACK_REWARD - 1.0,
            self.SUCCESS_REWARD + 1.0,
        )

    def get_reward(self, observations):
        reward = self.SLACK_REWARD

        current_measure = self._env.get_metrics()[self._reward_measure_name]

        reward += self._previous_measure - current_measure
        self._previous_measure = current_measure

        if self._episode_success():
            reward += self.SUCCESS_REWARD

        return reward

    def _episode_success(self):
        return self._env.get_metrics()[self._success_measure_name]

    def get_done(self, observations):
        done = False
        if self._env.episode_over or self._episode_success():
            done = True
        return done

    def get_info(self, observations):
        return self.habitat_env.get_metrics()


class MyHabitat(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)

        self.observation_space = gym.spaces.Box(0, 255, (84, 84, 3), np.uint8)
        self._action_space_dict = self.env.action_space
        self.action_space = gym.spaces.Discrete(4)

    def step(self, action):
        action_env = self._transform_action(action)
        obs, reward, done, info = self.env.step(action=action_env)
        obs = self._transform_obs(obs)
        return obs, reward, done, info

    def reset(self, **kwargs):
        obs = self.env.reset(**kwargs)
        obs = self._transform_obs(obs)
        return obs

    def _transform_obs(self, obs):
        img = obs['rgb']
        # crop and rescale 360,640 to 256,256
        img = center_crop(img, (360, 360))
        # img = scale_image(img, 256.0/360.0)
        img = scale_image(img, 84.0 / 360.0)
        return img

    def _transform_action(self, action_index):
        return {
            "action": list(self._action_space_dict.spaces.keys())[action_index],
            "action_args": list(self._action_space_dict.spaces.values())[action_index].sample(),
        }


import itertools
import numpy as np
from copy import deepcopy

import torch
import torch.nn as nn
import torch.optim as optim

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients


class DDPG(Algorithm):
    """
    Deep Deterministic Policy Gradient algorithm class.

    Algorithm class to execute DDPG, from Timothy P. Lillicrap et al.
    CONTINUOUS CONTROL WITH DEEP REINFORCEMENT LEARNING
    (https://arxiv.org/pdf/1509.02971.pdf). Algorithms are modules generally
    required by multiple workers, so DDPG.algo_factory(...) returns a function
    that can be passed on to workers to instantiate their own DDPG module.

    Parameters
    ----------
    device : torch.device
        CPU or specific GPU where class computations will take place.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    lr_pi : float
        Policy optimizer learning rate.
    lr_q : float
        Q-nets optimizer learning rate.
    gamma : float
        Discount factor parameter.
    polyak : float
        DDPG polyak averaging parameter.
    num_updates : int
        Num consecutive actor_critic updates before data collection continues.
    update_every : int
        Regularity of actor_critic updates in number environment steps.
    start_steps : int
        Num of initial random environment steps before learning starts.
    mini_batch_size : int
        Size of actor_critic update batches.
    target_update_interval : float
        regularity of target nets updates with respect to actor_critic Adam updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations in actor_critic updates.
    max_grad_norm : float
        Gradient clipping parameter.
    policy_loss_addons : list
        List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

    Examples
    --------
    >>> create_algo = DDPG.create_factory(
            lr_q=1e-4, lr_pi=1e-4, gamma=0.99, polyak=0.995,
            num_updates=50, update_every=50, test_every=5000, start_steps=20000,
            mini_batch_size=64, num_test_episodes=0, target_update_interval=1)
    """

    def __init__(self,
                 device,
                 envs,
                 actor,
                 lr_q=1e-4,
                 lr_pi=1e-4,
                 gamma=0.99,
                 polyak=0.995,
                 num_updates=1,
                 update_every=50,
                 test_every=1000,
                 max_grad_norm=0.5,
                 start_steps=20000,
                 mini_batch_size=64,
                 num_test_episodes=5,
                 target_update_interval=1,
                 policy_loss_addons=[]):

        # ---- General algo attributes ----------------------------------------

        # Discount factor
        self._gamma = gamma

        # Number of steps collected with initial random policy
        self._start_steps = int(start_steps)

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(1)  # Default to 1 for off-policy algorithms

        # Number of data samples collected between network update stages
        self._update_every = int(update_every)

        # Number mini batches per epoch
        self._num_mini_batch = int(num_updates)

        # Size of update mini batches
        self._mini_batch_size = int(mini_batch_size)

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = int(num_test_episodes)

        # ---- DDPG-specific attributes ----------------------------------------

        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.polyak = polyak
        self.device = device
        self.max_grad_norm = max_grad_norm
        self.target_update_interval = target_update_interval

        assert hasattr(self.actor, "q1"), "DDPG requires q critic (num_critics=1)"

        # Create target networks
        self.actor_targ = deepcopy(actor)

        # Freeze target networks with respect to optimizers (only update via polyak averaging)
        for p in self.actor_targ.parameters():
            p.requires_grad = False

        # List of parameters for both Q-networks
        q_params = itertools.chain(self.actor.q1.parameters())

        # List of parameters for both Q-networks
        p_params = itertools.chain(self.actor.policy_net.parameters())

        # ----- Policy Loss Addons --------------------------------------------

        # Sanity check, policy_loss_addons is a PolicyLossAddOn instance
        # or a list of PolicyLossAddOn instances
        assert isinstance(policy_loss_addons, (PolicyLossAddOn, list)),\
            "DDPG policy_loss_addons parameter should be a  PolicyLossAddOn instance " \
            "or a list of PolicyLossAddOn instances"
        if isinstance(policy_loss_addons, list):
            for addon in policy_loss_addons:
                assert isinstance(addon, PolicyLossAddOn), \
                    "DDPG policy_loss_addons parameter should be a  PolicyLossAddOn " \
                    "instance or a list of PolicyLossAddOn instances"
        else:
            policy_loss_addons = [policy_loss_addons]

        self.policy_loss_addons = policy_loss_addons
        for addon in self.policy_loss_addons:
            addon.setup(self.actor, self.device)

        # ----- Optimizers ----------------------------------------------------

        self.pi_optimizer = optim.Adam(p_params, lr=lr_pi)
        self.q_optimizer = optim.Adam(q_params, lr=lr_q)

    @classmethod
    def create_factory(cls,
                       lr_q=1e-3,
                       lr_pi=1e-4,
                       gamma=0.99,
                       polyak=0.995,
                       num_updates=50,
                       test_every=5000,
                       update_every=50,
                       start_steps=1000,
                       max_grad_norm=0.5,
                       mini_batch_size=64,
                       num_test_episodes=5,
                       target_update_interval=1.0,
                       policy_loss_addons=[]):
        """
        Returns a function to create new DDPG instances.

        Parameters
        ----------
        lr_pi : float
            Policy optimizer learning rate.
        lr_q : float
            Q-nets optimizer learning rate.
        gamma : float
            Discount factor parameter.
        polyak: float
            DDPG polyak averaging parameter.
        num_updates: int
            Num consecutive actor_critic updates before data collection continues.
        update_every: int
            Regularity of actor_critic updates in number environment steps.
        start_steps: int
            Num of initial random environment steps before learning starts.
        mini_batch_size: int
            Size of actor_critic update batches.
        target_update_interval: float
            regularity of target nets updates with respect to actor_critic Adam updates.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations in actor_critic updates.
        max_grad_norm : float
            Gradient clipping parameter.
        policy_loss_addons : list
            List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new DDPG class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr_q=lr_q,
                       lr_pi=lr_pi,
                       envs=envs,
                       actor=actor,
                       gamma=gamma,
                       device=device,
                       polyak=polyak,
                       test_every=test_every,
                       start_steps=start_steps,
                       num_updates=num_updates,
                       update_every=update_every,
                       max_grad_norm=max_grad_norm,
                       mini_batch_size=mini_batch_size,
                       num_test_episodes=num_test_episodes,
                       target_update_interval=target_update_interval,
                       policy_loss_addons=policy_loss_addons)

        return create_algo_instance, prl.DDPG

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return self._gamma

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        DDPG acting function.

        Parameters
        ----------
        obs : torch.tensor
            Current world observation
        rhs : torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done : torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action : torch.tensor
            Predicted next action.
        clipped_action : torch.tensor
            Predicted next action (clipped to be within action space).
        rhs : torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other : dict
            Additional DDPG predictions, which are not used in other algorithms.
        """

        with torch.no_grad():
            (action, clipped_action, logp_action, rhs,
             entropy_dist, dist) = self.actor.get_action(
                obs, rhs, done, deterministic=deterministic)

        return action, clipped_action, rhs, {}

    def compute_loss_q(self, data, n_step=1, per_weights=1):
        """
         Calculate DDPG Q-nets loss

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute TD3 losses.
        n_step : int or float
            Number of future steps used to computed the truncated n-step return value.
        per_weights :
            Prioritized Experience Replay (PER) important sampling weights or 1.0.

        Returns
        -------
        loss_q1 : torch.tensor
            Q1-net loss.
        loss_q2 : torch.tensor
            Q2-net loss.
        loss_q : torch.tensor
            Weighted average of loss_q1 and loss_q2.
        errors : torch.tensor
            TD errors.
         """

        o, rhs, d = data[prl.OBS], data[prl.RHS], data[prl.DONE]
        a, r = data[prl.ACT], data[prl.REW]
        o2, rhs2, d2 = data[prl.OBS2], data[prl.RHS2], data[prl.DONE2]

        # Q-values for all actions
        q = self.actor.get_q_scores(o, rhs, d, a).get("q1")

        # Bellman backup for Q functions
        with torch.no_grad():

            # Target actions come from *current* policy
            a2, _, _, _, _, dist = self.actor.get_action(o2, rhs2, d2)

            # Target Q-values
            q_targ = self.actor_targ.get_q_scores(o2, rhs2, d2, a2).get("q1")

            backup = r + (self.gamma ** n_step) * (1 - d2) * q_targ

        # MSE loss against Bellman backup
        loss_q = 0.5 * (((q - backup) ** 2) * per_weights).mean()

        # errors = (torch.min(q1, q2) - backup).abs()
        errors = (q - backup).abs()

        # reset Noise
        self.actor.policy_net.dist.noise.reset()

        return loss_q, errors

    def compute_loss_pi(self, data, per_weights=1):
        """
        Calculate DDPG policy loss.

        Parameters
        ----------
        data: dict
            Data batch dict containing all required tensors to compute DDPG losses.

        Returns
        -------
        loss_pi: torch.tensor
            DDPG policy loss.
        """

        o, rhs, d = data[prl.OBS], data[prl.RHS], data[prl.DONE]

        pi, _, _, _, _, dist = self.actor.get_action(o, rhs, d)
        q_pi = self.actor.get_q_scores(o, rhs, d, pi).get("q1")

        loss_pi = - (q_pi * per_weights).mean()

        # Extend policy loss with addons
        addons_info = {}
        for addon in self.policy_loss_addons:
            addon_loss, addons_info = addon.compute_loss_term(data, dist, addons_info)
            loss_pi += addon_loss

        info.update(addons_info)

        return loss_pi, addons_info

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        batch: dict
            data batch containing all required tensors to compute DDPG losses.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info: dict
            Dict containing current DDPG iteration information.
        """

        # Recurrent burn-in
        if self.actor.is_recurrent:
            batch = self.actor.burn_in_recurrent_states(batch)

        # PER
        per_weights = batch.pop("per_weights") if "per_weights" in batch else 1.0

        # N-step returns
        n_step = batch.pop("n_step") if "n_step" in batch else 1.0

        # First run one gradient descent step for Q1 and Q2
        loss_q, errors = self.compute_loss_q(batch, n_step, per_weights)
        self.q_optimizer.zero_grad()
        loss_q.backward(retain_graph=True)
        nn.utils.clip_grad_norm_(self.actor.q1.parameters(), self.max_grad_norm)
        q_grads = get_gradients(self.actor.q1, grads_to_cpu=grads_to_cpu)

        # Freeze Q-network so you don't waste computational effort
        # computing gradients for them during the policy learning step.
        for p in itertools.chain(self.actor.q1.parameters()):
            p.requires_grad = False

        # Next run one gradient descent step for pi.
        loss_pi, addons_info = self.compute_loss_pi(batch, per_weights)
        self.pi_optimizer.zero_grad()
        loss_pi.backward()
        nn.utils.clip_grad_norm_(self.actor.policy_net.parameters(), self.max_grad_norm)
        pi_grads = get_gradients(self.actor.policy_net, grads_to_cpu=grads_to_cpu)

        for p in itertools.chain(self.actor.q1.parameters()):
            p.requires_grad = True

        info = {
            "loss_pi": loss_pi.detach().item(),
            "loss_q1": loss_q.detach().item(),
        }

        if "per_weights" in batch:
            info.update({"errors": errors})

        grads = {"q_grads": q_grads, "pi_grads": pi_grads}

        return grads, info

    def update_target_networks(self):
        """Update actor critic target networks with polyak averaging"""
        if self.iter % self.target_update_interval == 0:
            with torch.no_grad():
                for p, p_targ in zip(self.actor.parameters(), self.actor_targ.parameters()):
                    p_targ.data.mul_(self.polyak)
                    p_targ.data.add_((1 - self.polyak) * p.data)

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients : list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.policy_net,
                gradients=gradients["pi_grads"], device=self.device)
            set_gradients(
                self.actor.q1,
                gradients=gradients["q_grads"], device=self.device)

        self.q_optimizer.step()
        self.pi_optimizer.step()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights. Update also target networks.

        Parameters
        ----------
        actor_weights : dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.pi_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
            for param_group in self.q_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import gym
from pytorchrl.agent.actors.feature_extractors.mlp import MLP
from pytorchrl.agent.actors.feature_extractors.cnn import CNN
from pytorchrl.agent.actors.feature_extractors.dictnet import DictNet
from pytorchrl.agent.actors.feature_extractors.fixup_cnn import FixupCNN
from pytorchrl.agent.actors.feature_extractors.embedding import Embedding


def get_feature_extractor(name):
    """Returns model class from name."""

    if name is None:
        return None
    elif name == "MLP":
        return MLP
    elif name == "CNN":
        return CNN
    elif name == "Fixup":
        return FixupCNN
    elif name == "DictNet":
        return DictNet
    elif name == "Embedding":
        return Embedding
    else:
        raise ValueError("Specified feature extractor model not found!")


def default_feature_extractor(input_space):
    """
    Returns the default net to use as a feature extractor
    given input_space.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    """

    if isinstance(input_space, gym.spaces.Dict):
        net = DictNet
    elif len(input_space.shape) <= 2:
        net = MLP
    elif len(input_space.shape) == 3:
        net = CNN
    else:
        raise NotImplementedError

    return net



import os
import gym
import glob
import random
import numpy as np
from PIL import Image

from animalai.envs.arena_config import ArenaConfig
from animalai.envs.gym.environment import ActionFlattener
from pytorchrl.envs.animal_olympics.utils import set_reward_arena


class RetroEnv(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)
        self.flattener = ActionFlattener([3, 3])
        self.action_space = self.flattener.action_space
        self.observation_space = gym.spaces.Box(0, 255, dtype=np.uint8, shape=(84, 84, 3))

    def step(self, action):
        obs, reward, done, info = self.env.step(action)  # non-retro
        visual_obs = self._preprocess_obs(obs)
        return visual_obs, reward, done, info

    def reset(self, **kwargs):
        obs = self.env.reset(**kwargs)
        visual_obs = self._preprocess_obs(obs)
        return visual_obs

    def _preprocess_obs(self,obs):
        visual_obs = obs
        visual_obs = self._preprocess_single(visual_obs)
        visual_obs = self._resize_observation(visual_obs)
        return visual_obs

    @staticmethod
    def _preprocess_single(single_visual_obs):
        return (255.0 * single_visual_obs).astype(np.uint8)

    @staticmethod
    def _resize_observation(observation):
        """
        Re-sizes visual observation to 84x84
        """
        obs_image = Image.fromarray(observation)
        obs_image = obs_image.resize((84, 84), Image.NEAREST)
        return np.array(obs_image)


# {0: [0, 0], 1: [0, 1], 2: [0, 2], 3: [1, 0], 4: [1, 1], 5: [1, 2], 6: [2, 0], 7: [2, 1], 8: [2, 2]}
class FilterActionEnv(gym.ActionWrapper):
    """
    An environment wrapper that limits the action space.
    """
    _ACTIONS = (0, 1, 2, 3, 4, 5, 6, 7, 8)
    # _ACTIONS = (0, 1, 2, 3, 4, 5, 6)

    def __init__(self, env):
        super().__init__(env)
        self.actions = self._ACTIONS
        self.action_space = gym.spaces.Discrete(len(self.actions))

    def step(self, action):
        action = int(action)
        action = self.flattener.lookup_action(action)  # convert to multi
        obs, reward, done, info = self.env.step(action)  # non-retro
        return obs, reward, done, info

    def action(self, act):
        return self.actions[act]


class LabAnimal(gym.Wrapper):
    def __init__(self, env, arenas_dir):
        gym.Wrapper.__init__(self, env)
        if os.path.isdir(arenas_dir):
            files = glob.glob("{}/*.yaml".format(arenas_dir))
        else:
            # assume is a pattern
            files = glob.glob(arenas_dir)

        self.env_list = [(f, ArenaConfig(f)) for f in files]
        self._arena_file = ''

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.steps += 1
        self.env_reward += reward
        info['arenas'] = self._arena_file  # for monitor
        info['max_reward'] = self.max_reward
        info['max_time'] = self.max_time
        info['ereward'] = self.env_reward
        return obs, reward, done, info

    def reset(self, **kwargs):
        self.steps = 0
        self.env_reward = 0
        self._arena_file, arena = random.choice(self.env_list)
        #        self.max_reward = analyze_arena(arenas)
        self.max_reward = set_reward_arena(arena, force_new_size=False)
        self.max_time = arena.arenas[0].t
        return self.env.reset(arenas_configurations=arena, **kwargs)


class RewardShaping(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        if (reward > -0.01) and (reward < 0):  # remove time negative reward
            reward = 0
        return obs, reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


import time
import torch
import numpy as np
from collections import defaultdict

import pytorchrl as prl
from pytorchrl.scheme.utils import check_message, pack
from pytorchrl.scheme.base.worker import Worker as W


class CWorker(W):
    """
    Worker class handling data collection.

    This class wraps an actor instance, a storage class instance and a
    train and a test vector environments. It collects data samples, sends
    them and and evaluates network versions.

    Parameters
    ----------
    index_worker : int
        Worker index.
    index_worker : int
        Index of gradient worker in charge of this data collection worker.
    algo_factory : func
        A function that creates an algorithm class.
    actor_factory : func
        A function that creates a policy.
    storage_factory : func
        A function that create a rollouts storage.
    fraction_samples : float
        Minimum fraction of samples required to stop if collection is
        synchronously coordinated and most workers have finished their
        collection task.
    compress_data_to_send : bool
        Whether or not to compress data before sending it to grad worker.
    train_envs_factory : func
        A function to create train environments.
    test_envs_factory : func
        A function to create test environments.
    initial_weights : ray object ID
        Initial model weights.
    device : str
        "cpu" or specific GPU "cuda:number`" to use for computation.
        
    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    fraction_samples : float
        Minimum fraction of samples required to stop if collection is
        synchronously coordinated and most workers have finished their
        collection task.
    device : torch.device
        CPU or specific GPU to use for computation.
    compress_data_to_send : bool
        Whether or not to compress data before sending it to grad worker.
    actor : Actor
        An actor class instance.
    algo : Algo
        An algorithm class instance.
    envs_train : VecEnv
        A VecEnv class instance with the train environments.
    envs_test : VecEnv
        A VecEnv class instance with the test environments.
    storage : Storage
        A Storage class instance.
    iter : int
         Number of times samples have been collected and sent.
    actor_version : int
        Number of times the current actor version been has been updated.
    update_every : int
        Number of data samples to collect between network update stages.
    obs : torch.tensor
        Latest train environment observation.
    rhs : torch.tensor
        Latest policy recurrent hidden state.
    done : torch.tensor
        Latest train environment done flag.
    """

    def __init__(self,
                 index_worker,
                 index_parent,
                 algo_factory,
                 actor_factory,
                 storage_factory,
                 fraction_samples=1.0,
                 compress_data_to_send=False,
                 train_envs_factory=lambda x, y, z: None,
                 test_envs_factory=lambda v, x, y, z: None,
                 initial_weights=None,
                 device=None):

        super(CWorker, self).__init__(index_worker)
        self.index_worker = index_worker
        self.fraction_samples = fraction_samples

        # Computation device
        dev = device or "cuda" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(dev)
        self.compress_data_to_send = compress_data_to_send

        # Create train environments
        self.envs_train = train_envs_factory(self.device, index_worker, index_parent)

        # Create test environments (if creation function available)
        self.envs_test = test_envs_factory(self.device, index_worker, index_parent, "test")

        # Create Actor Critic instance
        self.actor = actor_factory(self.device)

        # Create Algorithm instance
        self.algo = algo_factory(self.device, self.actor, self.envs_train)

        # Create Storage instance and set world initial state
        self.storage = storage_factory(self.device, self.actor, self.algo, self.envs_train)

        # Try to load weight from previous checkpoint to actor
        self.actor.try_load_from_checkpoint()

        # Define counters and other attributes
        self.iter, self.actor_version, self.samples_collected = 0, 0, 0
        self.update_every = self.algo.update_every or self.storage.max_size
        self.updates_per_iter = self.algo.num_mini_batch * self.algo.num_epochs

        if initial_weights:  # if remote worker

            # Set initial weights
            self.set_weights(initial_weights)

        if self.envs_train:
            # Define initial train states
            self.obs, self.rhs, self.done = self.actor.actor_initial_states(
                self.envs_train.reset())

            # Collect initial samples
            print("Collecting initial samples...")
            self.collect_train_data(self.algo.start_steps)

        # Print worker information
        self.print_worker_info()

    def collect_data(self, listen_to=[], data_to_cpu=True):
        """
        Perform a data collection operation, returning rollouts and
        other relevant information about the process.

        Parameters
        ----------
        listen_to : list
            List of keywords to listen to trigger early stopping during
            collection.

        Returns
        -------
        data : dict
            Collected train data samples.
        info : dict
            Additional relevant information about the collection operation.
        """

        # Define dictionary to record relevant information
        info = {prl.EPISODES: {}, prl.TIME: {}, prl.VERSION: {}}

        # Collect train data
        t = time.time()
        train_info = self.collect_train_data(listen_to=listen_to)
        col_time = time.time() - t

        # Add information to info dict
        info[prl.EPISODES] = train_info
        info[prl.TIME][prl.COLLECTION] = col_time
        info[prl.VERSION][prl.COLLECTION] = self.actor_version
        info[prl.NUMSAMPLES] = self.samples_collected

        # Get new data collected
        data = self.storage.get_all_buffer_data(data_to_cpu)

        # Evaluate current network on test environments
        if self.envs_test and self.algo.num_test_episodes > 0:

            # Actor model updates performed so far
            updates_so_far = self.iter * self.updates_per_iter

            # Actor model updates after using current train data
            updates_next_time = (self.iter + 1) * self.updates_per_iter

            # If self.algo.test_every actor model updates will be reached using
            # current train data, evaluate actor
            updates = np.arange(updates_so_far, updates_next_time)
            if (updates % self.algo.test_every == 0).sum() > 0:
                    test_perf = self.evaluate()
                    info[prl.EPISODES]["TestReward"] = test_perf

        # Encode data if self.compress_data_to_send is True
        data_to_send = pack((data, info)) if self.compress_data_to_send else (data, info)

        # Reset storage and number of collected samples
        self.storage.reset()
        self.samples_collected = 0

        # Update counter
        self.iter += 1

        return data_to_send

    def collect_train_data(self, num_steps=None, listen_to=[]):
        """
        Collect train data from interactions with the environments.

        Parameters
        ----------
        num_steps : int
            Target number of train environment steps to take.
        listen_to : list

        Returns
        -------
        col_time : float
            Time, in seconds, spent in this operation.
        train_perf : float
            Average accumulated reward over recent train episodes.
        """

        info = defaultdict(list)

        # Define number of collections steps to perform
        num_steps = int(num_steps) if num_steps is not None else int(self.update_every)
        min_steps = int(num_steps * self.fraction_samples)
        self.actor.eval()

        for step in range(num_steps):

            # Predict next action, next rnn hidden state and
            # algo-specific outputs
            act, clip_act, rhs2, algo_data = self.algo.acting_step(
                self.obs, self.rhs, self.done)

            # Interact with env with predicted action (clipped
            # within action space)
            obs2, reward, done2, episode_infos = self.envs_train.step(clip_act)

            # Define transition sample
            transition = prl.DataTransition(
                self.obs, self.rhs, self.done, act, reward,
                obs2, rhs2, done2, episode_infos)._asdict()
            transition.update(algo_data)

            # Store transition in buffer
            self.storage.insert_transition(transition)

            # Handle end of episode - collect episode info
            done_positions = torch.nonzero(done2, as_tuple=False)[:, 0].tolist()

            for i in done_positions:
                if "episode" in episode_infos[i]:  # gym envs should have it
                    for k, v in episode_infos[i]["episode"].items():
                        if isinstance(v, (float, int)):
                            if k == 'r':
                                k = "TrainReward"
                            info[k].append(v)

            # Update current world state
            self.obs, self.rhs, self.done = obs2, rhs2, done2

            # Keep track of num collected samples
            self.samples_collected += self.envs_train.num_envs

            # Check if stop message sent
            if self.is_remote:
                for l in listen_to:
                    if check_message(l) == b"stop" and step >= min_steps:
                        break

        # Average episodes infos
        info = {} if len(info) == 0 else {k: np.mean(v) for k, v in info.items()}

        self.actor.train()
        return info

    def evaluate(self):
        """
        Test current actor version in self.envs_test.

        Returns
        -------
        mean_test_perf : float
            Average accumulated reward over all tested episodes.
        """

        completed_episodes = []
        obs = self.envs_test.reset()
        rewards = np.zeros(obs.shape[0])
        obs, rhs, done = self.actor.actor_initial_states(obs)
        self.actor.eval()

        while len(completed_episodes) < self.algo.num_test_episodes:

            # Predict next action and rnn hidden state
            act, clip_act, rhs, _ = self.algo.acting_step(
                obs, rhs, done, deterministic=True)

            # Interact with env with predicted action (clipped
            # within action space)
            obs2, reward, done, _ = self.envs_test.step(clip_act)

            # Keep track of episode rewards and completed episodes
            rewards += reward.cpu().squeeze(-1).numpy()
            completed_episodes.extend(
                rewards[done.cpu().squeeze(-1).numpy() == 1.0].tolist())
            rewards[done.cpu().squeeze(-1).numpy() == 1.0] = 0.0

            obs = obs2

        self.actor.train()
        return np.mean(completed_episodes)

    def set_weights(self, actor_weights):
        """
        Update the worker actor version with provided weights.

        Parameters
        ----------
        actor_weights : dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor_version = actor_weights[prl.VERSION]
        self.actor.load_state_dict(actor_weights[prl.WEIGHTS])

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of self.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        new_parameter_value : float
            Algorithm new parameter value.
        """
        self.algo.update_algorithm_parameter(parameter_name, new_parameter_value)

    def update_storage_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of self.storage, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Storage attribute name
        new_parameter_value : float
            Storage new parameter value.
        """
        self.storage.update_storage_parameter(parameter_name, new_parameter_value)

    def replace_agent_component(self, component_name, new_component_factory):
        """
        If `component_name` is an attribute of c_worker, replaces it with
        the component created by `new_component_factory`.

        Parameters
        ----------
        component_name : str
            Worker component name
        new_component_factory : func
            Function to create an instance of the new component.
        """

        # TODO. get component name from component itself

        if hasattr(self, component_name):
            if component_name == prl.ALGORITHM:
                new_component_component = new_component_factory(
                    self.device, self.actor)
            elif component_name == prl.ENV_TRAIN:
                new_component_component = new_component_factory(
                    self.device, self.index_worker)
            elif component_name == prl.ENV_TEST:
                new_component_component = new_component_factory(
                    self.device, self.index_worker, "test")
            else:
                new_component_component = new_component_factory(self.device)
            setattr(self, component_name, new_component_component)

    def stop(self):
        """Stop all processes"""
        self.envs_train.close()
        try:
            self.envs_test.close()
        except Exception:
            pass



import torch
import torch.nn as nn
from torch.distributions.normal import Normal
import torch.nn.functional as F
from pytorchrl.agent.actors.feature_extractors.ensemble_layer import EnsembleFC

from pytorchrl.agent.actors.utils import init

LOG_STD_MAX = 2  # Maximum std allowed value. Used for clipping.
LOG_STD_MIN = -20  # Minimum std allowed value. Used for clipping.


class DiagGaussian(nn.Module):
    """
    Isotropic gaussian probability distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of dims in output space.
    predict_log_std : bool
        Whether to use a nn.linear layer to predict the output std.
    """

    def __init__(self, num_inputs, num_outputs, predict_log_std=False):
        super(DiagGaussian, self).__init__()

        self.predict_log_std = predict_log_std

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))
        self.mean = init_(nn.Linear(num_inputs, num_outputs))

        if predict_log_std:
            self.log_std = init_(nn.Linear(num_inputs, num_outputs))
        else:
            self.log_std = nn.Parameter(torch.zeros(num_outputs).unsqueeze(0))

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        predicted values (sampled and clipped), sampled log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample from predicted distribution or take the mode.

        Returns
        -------
        pred: torch.tensor
            Predicted value.
        clipped_pred: torch.tensor
            Predicted value (clipped to be within [-1, 1] range).
        logp : torch.tensor
            Log probability of `pred` according to the predicted distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted distribution.
        dist : torch.Distribution
            Action probability distribution.
        """
        # Predict distribution parameters
        mean = self.mean(x)
        logstd = self.log_std(x) if self.predict_log_std else torch.zeros(
            mean.size()).to(x.device) + self.log_std

        # logstd = torch.clamp(logstd, LOG_STD_MIN, LOG_STD_MAX)

        # Create distribution and sample
        dist = Normal(mean, logstd.exp())

        if deterministic:
            pred = mean
        else:
            pred = dist.sample()

        # Apply clipping to avoid being outside output space
        clipped_pred = torch.clamp(pred, -1, 1)

        # Action log probability
        logp = dist.log_prob(pred).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = dist.entropy().sum(-1).mean()

        return pred, clipped_pred, logp, entropy_dist, dist

    def evaluate_pred(self, x, pred):
        """
        Return log prob of `pred` under the distribution generated from
        x (obs features). Also return entropy of the generated distribution.

        Parameters
        ----------
        x : torch.tensor
            obs feature map obtained from a policy_net.
        pred : torch.tensor
            Prediction to evaluate.

        Returns
        -------
        logp : torch.tensor
            Log probability of `pred` according to the predicted distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted distribution.
        dist : torch.Distribution
            Action probability distribution.
        """

        # Predict distribution parameters
        mean = self.mean(x)
        logstd = self.log_std(x) if self.predict_log_std else torch.zeros(
            mean.size()).to(x.device) + self.log_std

        # logstd = torch.clamp(logstd, LOG_STD_MIN, LOG_STD_MAX)

        # Create distribution
        dist = Normal(mean, logstd.exp())

        # Evaluate log prob under dist
        logp = dist.log_prob(pred).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = dist.entropy().sum(-1).mean()

        return logp, entropy_dist, dist

class DiagGaussianEnsemble(nn.Module):
    """Ensemble gaussian output layer 
    
        Parameters
        ----------
        num_inputs: int
            Size of input feature maps.
        num_outputs: int
            Output size of the gaussian layer.
        ensemble_size: int
            Ensemble size in the output layer.
    """
    def __init__(self, num_inputs: int, num_outputs: int, ensemble_size: int)-> None:
        super(DiagGaussianEnsemble, self).__init__()

        self.num_outputs = num_outputs
        self.ensemble_size = ensemble_size
        self.mu = EnsembleFC(in_features=num_inputs, out_features=num_outputs, ensemble_size=ensemble_size)

    def forward(self, x: torch.Tensor):
        """Forward pass"""
        return self.mu(x)



import os
import obstacle_tower_env
from obstacle_tower_env import ObstacleTowerEnv
from pytorchrl.envs.common import FrameStack, FrameSkip, DelayedReward
from pytorchrl.envs.obstacle_tower.wrappers import (
    ReducedActionEnv, BasicObstacleEnv, RewardShapeObstacleEnv, BasicObstacleEnvTest)


def obstacle_train_env_factory(
        index_col_worker, index_grad_worker, index_env=0, frame_skip=0, frame_stack=1, min_floor=0,
        max_floor=50, reduced_actions=True, num_actions=6, reward_shape=True, exe_path=None, reward_delay=1,
        realtime=False, seed_list=[], id_offset=1, timeout_wait=180):
    """
    Create train Obstacle Tower Unity3D environment.
    Useful info_keywords 'floor', 'start', 'seed'.
    Parameters
    ----------
     index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    min_floor : int
        Minimum floor the agent can be spawned in.
    max_floor : int
        Maximum floor the agent can be spawned in.
    reduced_actions : bool
        Whether or not to use the action wrapper to reduce the number of available actions.
    num_actions : int
        Size of the reduced action space.
    reward_shape : bool
        Whether or not to use the reward shape wrapper.
    exe_path : str
        Path to obstacle environment executable.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    realtime : bool
        Whether or not to render the environment frames in real time.
    seed_list : list
        List of environment seeds to use.
    id_offset : int
        offset added to worker_id to avoid collisions with other runs in the same machine.
    timeout_wait : int
        Time for python interface to wait for environment to connect (in seconds).

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    if exe_path:
        exe = exe_path
    else:
        exe = os.path.join(os.path.dirname(
            obstacle_tower_env.__file__), 'ObstacleTower/obstacletower')

    id = id_offset + index_grad_worker * 1000 + 100 * index_col_worker + index_env
    env = ObstacleTowerEnv(
        environment_filename=exe, retro=True, worker_id=id,
        greyscale=False, timeout_wait=timeout_wait, realtime_mode=realtime)

    if reduced_actions:
        env = ReducedActionEnv(env, num_actions=num_actions)

    env = BasicObstacleEnv(env, min_floor=min_floor, max_floor=max_floor, seed_list=seed_list)

    if reward_shape:
        env = RewardShapeObstacleEnv(env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env


def obstacle_test_env_factory(
        index_col_worker, index_grad_worker, index_env=0, frame_skip=0, frame_stack=1, realtime=False,
        min_floor=0, max_floor=50, reduced_actions=True, num_actions=6, exe_path=None, reward_delay=1,
        id_offset=1, timeout_wait=180):
    """
    Create test Obstacle Tower Unity3D environment.
    Useful info_keywords 'floor', 'start', 'seed'.
    Parameters
    ----------
    index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    min_floor : int
        Minimum floor the agent can be spawned in.
    max_floor : int
        Maximum floor the agent can be spawned in.
    reduced_actions : bool
        Whether or not to use the action wrapper to reduce the number of available actions.
    num_actions : int
        Size of the reduced action space.
    exe_path : str
        Path to obstacle environment executable.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    realtime : bool
        Whether or not to render the environment frames in real time.
    id_offset : int
        offset added to worker_id to avoid collisions with other runs in the same machine.
    timeout_wait : int
        Time for python interface to wait for environment to connect (in seconds).

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    if exe_path:
        exe = exe_path
    else:
        exe = os.path.join(os.path.dirname(
            obstacle_tower_env.__file__), 'ObstacleTower/obstacletower')

    id = id_offset + index_grad_worker * 1000 + 100 * index_col_worker + index_env
    env = ObstacleTowerEnv(
        environment_filename=exe, retro=True, worker_id=id,
        greyscale=False, timeout_wait=timeout_wait, realtime_mode=realtime)

    if reduced_actions:
        env = ReducedActionEnv(env, num_actions=num_actions)

    env = BasicObstacleEnvTest(env, max_floor=max_floor, min_floor=min_floor)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env



import gym
import torch
import torch.nn as nn
import numpy as np
from collections import OrderedDict

import pytorchrl as prl
from pytorchrl.agent.actors.base import Actor
from pytorchrl.agent.actors.distributions import get_dist
from pytorchrl.agent.actors.utils import Scale, Unscale, init
from pytorchrl.agent.actors.feature_extractors import MLP, default_feature_extractor


class OffPolicyActor(Actor):
    """
    Actor critic class for Off-Policy algorithms.

    It contains a policy network (actor) to predict next actions and one or two
    Q networks.

    Parameters
    ----------
    device: torch.device
        CPU or specific GPU where class computations will take place.
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    algorithm_name : str
        Name of the RL algorithm used for learning.
    checkpoint : str
        Path to a previously trained Actor checkpoint to be loaded.
    noise : str
        Type of exploration noise that will be added to the deterministic actions.
    obs_feature_extractor : nn.Module
        PyTorch nn.Module to extract features from observation in all networks.
    obs_feature_extractor_kwargs : dict
        Keyword arguments for the obs extractor network.
    act_feature_extractor : nn.Module
        PyTorch nn.Module to extract features from actions in all networks.
    act_feature_extractor_kwargs : dict
        Keyword arguments for the act extractor network.
    common_feature_extractor : nn.Module
        PyTorch nn.Module to extract joint features from the concatenation of
        action and observation features.
    common_feature_extractor_kwargs : dict
        Keyword arguments for the common extractor network.
    recurrent_net : bool
        Whether to use a RNNs as feature extractors.
    sequence_overlap : float
        From 0.0 to 1.0, how much consecutive rollout sequences will overlap.
    recurrent_net_kwargs : dict
        Keyword arguments for the memory network.
    num_critics : int
        Number of Q networks to be instantiated.

    Examples
    --------
    """
    def __init__(self,
                 device,
                 input_space,
                 action_space,
                 algorithm_name,
                 noise=None,
                 checkpoint=None,
                 sequence_overlap=0.5,
                 recurrent_net=None,
                 recurrent_net_kwargs={},
                 obs_feature_extractor=None,
                 obs_feature_extractor_kwargs={},
                 act_feature_extractor=None,
                 act_feature_extractor_kwargs={},
                 common_feature_extractor=MLP,
                 common_feature_extractor_kwargs={},
                 num_critics=2):

        super(OffPolicyActor, self).__init__(
            device=device,
            checkpoint=checkpoint,
            input_space=input_space,
            action_space=action_space)

        self.noise = noise
        self.algorithm_name = algorithm_name
        self.input_space = input_space
        self.action_space = action_space
        self.act_feature_extractor = act_feature_extractor
        self.act_feature_extractor_kwargs = act_feature_extractor_kwargs
        self.obs_feature_extractor = obs_feature_extractor
        self.obs_feature_extractor_kwargs = obs_feature_extractor_kwargs
        self.common_feature_extractor = common_feature_extractor
        self.common_feature_extractor_kwargs = common_feature_extractor_kwargs
        self.recurrent_net = recurrent_net
        self.recurrent_net_kwargs = recurrent_net_kwargs
        self.sequence_overlap = np.clip(sequence_overlap, 0.0, 1.0)
        self.num_critics = num_critics
        self.deterministic = algorithm_name in [prl.DDPG, prl.TD3]

        # ----- Policy Network ----------------------------------------------------

        self.create_policy("policy_net")

        # ----- Q Networks ----------------------------------------------------

        for i in range(num_critics):
            self.create_critic("q{}".format(i + 1))

    @classmethod
    def create_factory(
            cls,
            input_space,
            action_space,
            algorithm_name,
            noise=None,
            restart_model=None,
            sequence_overlap=0.5,
            recurrent_net_kwargs={},
            recurrent_net=None,
            obs_feature_extractor=None,
            obs_feature_extractor_kwargs={},
            act_feature_extractor=None,
            act_feature_extractor_kwargs={},
            common_feature_extractor=MLP,
            common_feature_extractor_kwargs={},
            num_critics=2

    ):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        algorithm_name : str
            Name of the RL algorithm used for learning.
        noise : str
            Type of exploration noise that will be added to the deterministic actions.
        obs_feature_extractor : nn.Module
            PyTorch nn.Module to extract features from observation in all networks.
        obs_feature_extractor_kwargs : dict
            Keyword arguments for the obs extractor network.
        act_feature_extractor : nn.Module
            PyTorch nn.Module to extract features from actions in all networks.
        act_feature_extractor_kwargs : dict
            Keyword arguments for the act extractor network.
        common_feature_extractor : nn.Module
            PyTorch nn.Module to extract joint features from the concatenation of
            action and observation features.
        common_feature_extractor_kwargs : dict
            Keyword arguments for the common extractor network.
        recurrent_net : bool
            Whether to use a RNNs as feature extractors.
        sequence_overlap : float
            From 0.0 to 1.0, how much consecutive rollout sequences will overlap.
        recurrent_net_kwargs : dict
            Keyword arguments for the memory network.
        num_critics : int
            Number of Q networks to be instantiated.
        restart_model : str
            Path to a previously trained Actor checkpoint to be loaded.

        Returns
        -------
        create_actor_instance : func
            creates a new OffPolicyActor class instance.
        """

        def create_actor_instance(device):
            """Create and return an actor critic instance."""
            policy = cls(noise=noise,
                         device=device,
                         input_space=input_space,
                         action_space=action_space,
                         algorithm_name=algorithm_name,
                         checkpoint=restart_model,
                         sequence_overlap=sequence_overlap,
                         recurrent_net_kwargs=recurrent_net_kwargs,
                         recurrent_net=recurrent_net,
                         obs_feature_extractor=obs_feature_extractor,
                         obs_feature_extractor_kwargs=obs_feature_extractor_kwargs,
                         act_feature_extractor=act_feature_extractor,
                         act_feature_extractor_kwargs=act_feature_extractor_kwargs,
                         common_feature_extractor=common_feature_extractor,
                         common_feature_extractor_kwargs=common_feature_extractor_kwargs,
                         num_critics=num_critics)
            policy.to(device)

            try:
                policy.try_load_from_checkpoint()
            except RuntimeError:
                pass

            return policy

        return create_actor_instance

    @property
    def is_recurrent(self):
        """Returns True if the actor network are recurrent."""
        return self.recurrent_net

    @property
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        return self.recurrent_size

    def actor_initial_states(self, obs):
        """
        Returns all actor inputs required to predict initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : dict
            Initial recurrent hidden state (will contain zeroes).
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """

        if isinstance(obs, dict):
            num_proc = list(obs.values())[0].size(0)
            dev = list(obs.values())[0].device
        else:
            num_proc = obs.size(0)
            dev = obs.device

        done = torch.zeros(num_proc, 1).to(dev)
        try:
            rhs_act = self.policy_net.memory_net.get_initial_recurrent_state(num_proc).to(dev)
        except Exception:
            rhs_act = torch.zeros(num_proc, self.recurrent_hidden_state_size).to(dev)

        rhs = {"rhs_act": rhs_act}
        rhs.update({"rhs_q{}".format(i + 1): rhs_act.clone() for i in range(self.num_critics)})

        return obs, rhs, done

    def burn_in_recurrent_states(self, data_batch):
        """
        Applies a recurrent burn-in phase to data_batch as described in
        (https://openreview.net/pdf?id=r1lyTjAqYX). Initial B steps are used
        to compute on-policy recurrent hidden states. data_batch is then
        updated, discarding B first steps in all tensors.

        Parameters
        ----------
        data_batch : dict
            data batch containing all required tensors to compute Algorithm loss.

        Returns
        -------
        data_batch : dict
            Updated data batch after burn-in phase.
        """

        # (T, N, -1) tensors that have been flatten to (T * N, -1)
        N = data_batch[prl.RHS]["rhs_act"].shape[0]  # number of sequences
        T = int(data_batch[prl.DONE].shape[0] / N)  # sequence lengths
        B = int(self.sequence_overlap * T)  # sequence burn-in length

        if B == 0:
            return data_batch

        # Split tensors into burn-in and no-burn-in
        chunk_sizes = [B, T - B] * N
        burn_in_data = {k: {} for k in data_batch}
        non_burn_in_data = {k: {} for k in data_batch}
        for k, v in data_batch.items():

            if k in (prl.RHS, prl.RHS2):
                burn_in_data[k] = v
                continue
            if not isinstance(v, (torch.Tensor, dict)):
                non_burn_in_data[k] = v
                continue
            if isinstance(v, dict):
                for x, y in v.items():
                    if not isinstance(y, torch.Tensor):
                        non_burn_in_data[k][x] = v
                        continue
                    sequence_slices = torch.split(y, chunk_sizes)
                    burn_in_data[k][x] = torch.cat(sequence_slices[0::2])
                    non_burn_in_data[k][x] = torch.cat(sequence_slices[1::2])
            else:
                sequence_slices = torch.split(v, chunk_sizes)
                burn_in_data[k] = torch.cat(sequence_slices[0::2])
                non_burn_in_data[k] = torch.cat(sequence_slices[1::2])

        # Do burn-in
        with torch.no_grad():

            act, _, _, rhs, _, _ = self.get_action(
                burn_in_data[prl.OBS], burn_in_data[prl.RHS], burn_in_data[prl.DONE])
            act2, _, _, rhs2, _, _ = self.get_action(
                burn_in_data[prl.OBS2], burn_in_data[prl.RHS2], burn_in_data[prl.DONE2])

            rhs = self.get_q_scores(
                burn_in_data[prl.OBS], rhs, burn_in_data[prl.DONE], act).get("rhs")
            rhs2 = self.get_q_scores(
                burn_in_data[prl.OBS2], rhs2, burn_in_data[prl.DONE2], act2).get("rhs")

            for k in rhs:
                rhs[k] = rhs[k].detach()
            for k in rhs2:
                rhs2[k] = rhs2[k].detach()

            non_burn_in_data[prl.RHS] = rhs
            non_burn_in_data[prl.RHS2] = rhs2

        return non_burn_in_data

    def get_action(self, obs, rhs, done, deterministic=False):
        """
        Predict and return next action, along with other information.

        Parameters
        ----------
        obs : torch.tensor
            Current environment observation.
        rhs : dict
            Current recurrent hidden states.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        action : torch.tensor
            Next action sampled.
        clipped_action : torch.tensor
            Next action sampled, but clipped to be within the env action space.
        logp_action : torch.tensor
            Log probability of `action` within the predicted action distribution.
        rhs : dict
            Updated recurrent hidden states.
        entropy_dist : torch.tensor
            Entropy of the predicted action distribution.
        dist : torch.Distribution
            Predicted probability distribution over next action.
        """

        x = self.policy_net.common_feature_extractor(self.policy_net.obs_feature_extractor(obs))

        if self.recurrent_net:
            x, rhs["rhs_act"] = self.policy_net.memory_net(x, rhs["rhs_act"], done)

        (action, clipped_action, logp_action, entropy_dist, dist) = self.policy_net.dist(
            x, deterministic=deterministic)

        if self.unscale:
            action = self.unscale(action)
            clipped_action = self.unscale(clipped_action)

        return action, clipped_action, logp_action, rhs, entropy_dist, dist

    def evaluate_actions(self, obs, rhs, done, action):
        """
        Evaluate log likelihood of action given obs and the current
        policy network. Returns also entropy distribution.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : dict
            Recurrent hidden states.
        done : torch.tensor
            Done tensor, indicating if episode has finished.
        action : torch.tensor
            Evaluated action.

        Returns
        -------
        logp_action : torch.tensor
            Log probability of `action` according to the action distribution
            predicted with current version of the policy_net.
        entropy_dist : torch.tensor
            Entropy of the action distribution predicted with current version
            of the policy_net.
        dist : torch.Distribution
            Predicted probability distribution over next action.
        """

        if self.scale:
            action = self.scale(action)

        features = self.policy_net.common_feature_extractor(self.policy_net.obs_feature_extractor(obs))

        if self.recurrent_net:
            features, rhs["rhs_act"] = self.policy_net.memory_net(features, rhs["rhs_act"], done)

        logp_action, entropy_dist, dist = self.policy_net.dist.evaluate_pred(features, action)

        return logp_action, entropy_dist, dist

    def get_q_scores(self, obs, rhs, done, actions=None):
        """
        Return Q scores of the given observations and actions.

        Parameters
        ----------
        obs : torch.tensor
            Environment observation.
        rhs : dict
            Current recurrent hidden states.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        actions : torch.tensor
             Evaluated actions.

        Returns
        -------
        output : dict
            Dict containing value prediction from each critic under keys "q1", "q2", etc
            as well as the recurrent hidden states under the key "rhs".
        """

        outputs = {}
        for i in range(self.num_critics):
            q = getattr(self, "q{}".format(i + 1))
            features = q.obs_feature_extractor(obs)

            if actions is not None:
                act_features = q.act_feature_extractor(actions)
                features = torch.cat([features, act_features], -1)
            features = q.common_feature_extractor(features)

            if self.recurrent_net:
                features, rhs["rhs_q{}".format(1 + 1)] = q.memory_net(
                    features, rhs["rhs_q{}".format(i + 1)], done)

            q_scores = q.predictor(features)
            outputs["q{}".format(i + 1)] = q_scores

        outputs["rhs"] = rhs
        return outputs

    def create_critic(self, name):
        """
        Create a critic q network and define it as class attribute under the name `name`.
        This actor defines defines q networks as:

            obs_feature_extractor
        q =                       + common_feature_extractor + memory_net + q_prediction_layer
            act_feature_extractor

        Parameters
        ----------
        name : str
            Critic network name.
        """

        # ---- 1. Define action feature extractor -----------------------------

        act_extractor = self.act_feature_extractor or nn.Identity
        q_act_feature_extractor = act_extractor(
            self.action_space, **self.act_feature_extractor_kwargs)

        # ---- 2. Define obs feature extractor -----------------------------

        obs_extractor = self.obs_feature_extractor or nn.Identity
        q_obs_feature_extractor = obs_extractor(
            self.input_space, **self.obs_feature_extractor_kwargs)
        obs_feature_size = q_obs_feature_extractor(
            torch.zeros(1, *self.input_space.shape)).shape[-1]

        # ---- 3. Define shared feature extractor -----------------------------

        if isinstance(self.action_space, gym.spaces.Discrete):
            act_feature_size = 0
            q_outputs = self.action_space.n

        elif isinstance(self.action_space, gym.spaces.Box):
            act_feature_size = q_act_feature_extractor(
                torch.zeros(1, *self.action_space.shape)).shape[-1] if self.act_feature_extractor \
                else self.action_space.shape[-1]
            q_outputs = 1

        else:
            raise NotImplementedError

        feature_size = obs_feature_size + act_feature_size
        q_common_feature_extractor = self.common_feature_extractor(
            feature_size, **self.common_feature_extractor_kwargs)

        # ---- 4. Define memory network ---------------------------------------

        feature_size = q_common_feature_extractor(torch.zeros(1, feature_size)).shape[-1]
        q_memory_net = self.recurrent_net(feature_size, **self.recurrent_net_kwargs) if\
            self.recurrent_net else nn.Identity()
        feature_size = q_memory_net.recurrent_hidden_state_size if self.recurrent_net\
            else feature_size

        # ---- 5. Define prediction layer -------------------------------------

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))
        q_predictor = init_(nn.Linear(feature_size, q_outputs))

        # ---- 6. Concatenate all q1 net modules ------------------------------

        q_net = nn.Sequential(OrderedDict([
            ('obs_feature_extractor', q_obs_feature_extractor),
            ('act_feature_extractor', q_act_feature_extractor),
            ('common_feature_extractor', q_common_feature_extractor),
            ('memory_net', q_memory_net),
            ("predictor", q_predictor),
        ]))

        setattr(self, name, q_net)

    def create_policy(self, name):
        """
        Create a policy network and define it as class attribute under the name `name`.
        This actor defines policy network as:

        policy = obs_feature_extractor + common_feature_extractor + memory_net + action distribution

        Parameters
        ----------
        name : str
            Policy network name.
        """

        # ---- 1. Define Obs feature extractor --------------------------------

        if self.obs_feature_extractor:
            self.obs_feature_extractor = self.obs_feature_extractor
        else:
            self.obs_feature_extractor = default_feature_extractor(self.input_space)

        obs_extractor = self.obs_feature_extractor or nn.Identity
        policy_obs_feature_extractor = obs_extractor(
            self.input_space, **self.obs_feature_extractor_kwargs)

        # ---- 2. Define Common feature extractor -----------------------------

        feature_size = policy_obs_feature_extractor(torch.zeros(1, *self.input_space.shape)).shape[-1]

        policy_common_feature_extractor = self.common_feature_extractor(
            feature_size, **self.common_feature_extractor_kwargs)

        # ---- 3. Define memory network  --------------------------------------

        feature_size = policy_common_feature_extractor(torch.zeros(1, feature_size)).shape[-1]

        if self.recurrent_net:
            policy_memory_net = self.recurrent_net(feature_size, **self.recurrent_net_kwargs)
            self.recurrent_size = policy_memory_net.recurrent_hidden_state_size
        else:
            policy_memory_net = nn.Identity()
            self.recurrent_size = feature_size

        # ---- 4. Define action distribution ----------------------------------

        if isinstance(self.action_space, gym.spaces.Discrete):
            dist = get_dist("Categorical")(self.recurrent_size, self.action_space.n)
            self.scale = None
            self.unscale = None

        elif isinstance(self.action_space, gym.spaces.Box) and not self.deterministic:
            if self.algorithm_name in [prl.SAC]:
                dist = get_dist("SquashedGaussian")(self.recurrent_size, self.action_space.shape[0])
            else:
                dist = get_dist("Gaussian")(self.recurrent_size, self.action_space.shape[0])
            self.scale = Scale(self.action_space)
            self.unscale = Unscale(self.action_space)

        elif isinstance(self.action_space, gym.spaces.Box) and self.deterministic:
            dist = get_dist("Deterministic")(
                self.recurrent_size, self.action_space.shape[0], noise=self.noise)
            self.scale = Scale(self.action_space)
            self.unscale = Unscale(self.action_space)
        else:
            raise NotImplementedError

        # ---- 5. Concatenate all policy modules ------------------------------

        policy_net = nn.Sequential(OrderedDict([
            ('obs_feature_extractor', policy_obs_feature_extractor),
            ('common_feature_extractor', policy_common_feature_extractor),
            ('memory_net', policy_memory_net), ('dist', dist),
        ]))

        setattr(self, name, policy_net)



import torch
import math

"""Good Resource of reward functions: https://arxiv.org/pdf/1907.02057.pdf 
    including gym envrionemnts as well as pybullet environments
"""


def cartpole(state: torch.Tensor, action: torch.Tensor, next_state: torch.Tensor) -> torch.Tensor:
    """
    Based on https://arxiv.org/pdf/1907.02057.pdf
    reward = cos(θ_t) - 0.01x²
    """
    x, x_dot, theta, theta_dot = state[:, 0], state[:, 1], state[:, 2], state[:, 3]

    reward = torch.cos(theta)[:, None] - 0.01 * x[:, None] ** 2
    return reward


def pendulum(state: torch.Tensor, action: torch.Tensor, next_state: torch.Tensor) -> torch.Tensor:
    # Original env reward function seems not to work!
    # max_torque = 2.0
    # th = torch.acos(state[:, 0][:, None])
    # thdot = state[:, 2][:, None]
    # action = torch.clamp(action, -max_torque, max_torque)

    # reward = angle_normalize(th) ** 2 + 0.1 * thdot ** 2 + 0.001 * (action ** 2)

    # reward function from Paper: https://arxiv.org/pdf/1907.02057.pdf
    cos_theta, sin_theta, theta_dot = state[:, 0], state[:, 1], state[:, 2]
    reward = - cos_theta[:, None] - 0.1 * sin_theta[:, None] - 0.1 * theta_dot[:, None] ** 2 - 0.001 * action ** 2
    return reward


def angle_normalize(x: torch.Tensor):
    return ((x + math.pi) % (2 * math.pi)) - math.pi


def inverted_pendulum_mujoco(state: torch.Tensor, action: torch.Tensor, next_state: torch.Tensor) -> torch.Tensor:
    """
    Env info: https://github.com/openai/gym/blob/master/gym/envs/mujoco/inverted_pendulum.py
    Reward function based on: https://arxiv.org/pdf/1907.02057.pdf
    
    reward = - theta², where theta = state[1]
    
    """
    reward = - state[:, 1][:, None] ** 2
    return reward


def halfcheetah_mujoco(state: torch.Tensor, action: torch.Tensor, next_state: torch.Tensor) -> torch.Tensor:
    """
    First 8 values in the state are position data
    other 9 are position velocities (x,y,z) and rest angular
    -> idx 8 is x_velocitiy
    """
    x_velocities = state[:, 8]
    action_penalty = - 0.1 * (torch.sum(action ** 2, axis=1))
    reward = (x_velocities + action_penalty)[:, None]
    assert reward.shape == (state.shape[0], 1)
    return reward



import abc
import pandas as pd
from reinvent_chemistry.conversions import Conversions
from pytorchrl.envs.generative_chemistry.diversity_filter.diversity_filter_memory import DiversityFilterMemory


class BaseDiversityFilter(abc.ABC):

    @abc.abstractmethod
    def __init__(self, minscore=0.4, bucket_size=25, minsimilarity=0.0):
        self.minscore = minscore
        self.bucket_size = bucket_size
        self.minsimilarity = minsimilarity
        self._chemistry = Conversions()
        self._diversity_filter_memory = DiversityFilterMemory()

    @abc.abstractmethod
    def update_score(self, score, smile):
        raise NotImplementedError("The method 'evaluate' is not implemented!")

    def get_memory_as_dataframe(self) -> pd.DataFrame:
        return self._diversity_filter_memory.get_memory()

    def set_memory_from_dataframe(self, memory: pd.DataFrame):
        self._diversity_filter_memory.set_memory(memory)

    def number_of_smiles_in_memory(self) -> int:
        return self._diversity_filter_memory.number_of_smiles()

    def _smiles_exists(self, smile):
        return self._diversity_filter_memory.smiles_exists(smile)

    def _add_to_memory(self, smile):
        self._diversity_filter_memory.update(smile)




from pytorchrl.agent.storages.off_policy.replay_buffer import ReplayBuffer
from pytorchrl.agent.storages.off_policy.nstep_buffer import NStepReplayBuffer
from pytorchrl.agent.storages.off_policy.per_buffer import PERBuffer
from pytorchrl.agent.storages.off_policy.ere_buffer import EREBuffer
from pytorchrl.agent.storages.off_policy.her_buffer import HERBuffer

from pytorchrl.agent.storages.on_policy.gae_buffer import GAEBuffer
from pytorchrl.agent.storages.on_policy.ppod_buffer import PPODBuffer
from pytorchrl.agent.storages.on_policy.ppod2_buffer import PPOD2Buffer
from pytorchrl.agent.storages.on_policy.ppod2rebel_buffer import PPOD2RebelBuffer
from pytorchrl.agent.storages.on_policy.vtrace_buffer import VTraceBuffer
from pytorchrl.agent.storages.on_policy.vanilla_on_policy_buffer import VanillaOnPolicyBuffer

from pytorchrl.agent.storages.model_based.mb_buffer import MBReplayBuffer



from pytorchrl.agent.actors.off_policy_actor import OffPolicyActor
from pytorchrl.agent.actors.on_policy_actor import OnPolicyActor
from pytorchrl.agent.actors.model_based_planner_actor import ModelBasedPlannerActor
from pytorchrl.agent.actors.feature_extractors import get_feature_extractor, default_feature_extractor
from pytorchrl.agent.actors.memory_networks import get_memory_network
